# "title/ticket.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki/Ticket for details about the ticket format

import io
import binascii
import hashlib
from dataclasses import dataclass as _dataclass
from .crypto import decrypt_title_key
from typing import List
from .util import title_ver_standard_to_dec


@_dataclass
class _TitleLimit:
    """
    A TitleLimit object that contains the type of restriction and the limit. The limit type can be one of the following:
    0 = None, 1 = Time Limit, 3 = None, or 4 = Launch Count. The maximum usage is then either the time in minutes the
    title can be played or the maximum number of launches allowed for that title, based on the type of limit applied.
    Private class used only by the Ticket class.

    Attributes
    ----------
    limit_type : int
        The type of play limit applied. 0 and 3 are none, 1 is a time limit, and 4 is a launch count limit.
    maximum_usage : int
        The maximum value for the type of play limit applied.
    """
    # The type of play limit applied.
    limit_type: int
    # The maximum value of the limit applied.
    maximum_usage: int


class Ticket:
    """
    A Ticket object that allows for either loading and editing an existing Ticket or creating one manually if desired.

    Attributes
    ----------
    is_dev : bool
        Whether this Ticket is signed for development or not, and whether the Title Key is encrypted for development
        or not.
    signature : bytes
        The signature applied to the ticket.
    ticket_version : int
        The version of the ticket.
    title_key_enc : bytes
        The Title Key contained in the ticket, in encrypted form.
    ticket_id : bytes
        The unique ID of this ticket, used for console-specific title installations.
    console_id : int
        The unique ID of the console this ticket was designed for, if this is a console-specific ticket.
    title_version : int
        The version of the title this ticket was designed for.
    common_key_index : int
        The index of the common key required to decrypt this ticket's Title Key.
    """
    def __init__(self):
        # If this is a dev ticket
        self.is_dev: bool = False  # Defaults to false, set to true during load if this ticket is using dev certs.
        # Signature blob header
        self.signature_type: bytes = b''  # Type of signature, always 0x10001 for RSA-2048
        self.signature: bytes = b''  # Actual signature data
        # v0 ticket data
        self.signature_issuer: str = ""  # Who issued the signature for the ticket
        self.ecdh_data: bytes = b''  # Involved in created one-time keys for console-specific title installs.
        self.ticket_version: int = 0  # The version of the current ticket file.
        self.title_key_enc: bytes = b''  # The title key of the ticket's respective title, encrypted by a common key.
        self.ticket_id: bytes = b''  # Used as the IV when decrypting the title key for console-specific title installs.
        self.console_id: int = 0  # ID of the console that the ticket was issued for.
        self.title_id: bytes = b''  # TID/IV used for AES-CBC encryption.
        self.unknown1: bytes = b''  # Some unknown data, not always the same so reading it just in case.
        self.title_version: int = 0  # Version of the ticket's associated title.
        self.permitted_titles: bytes = b''  # Permitted titles mask
        # "Permit mask. The current disc title is ANDed with the inverse of this mask to see if the result matches the
        # Permitted Titles Mask."
        self.permit_mask: bytes = b''
        self.title_export_allowed: int = 0  # Whether title export is allowed with a PRNG key or not.
        self.common_key_index: int = 0  # Which common key should be used. 0 = Common Key, 1 = Korean Key, 2 = vWii Key
        self.unknown2: bytes = b''  # More unknown data. Varies for VC/non-VC titles so reading it to ensure it matches.
        self.content_access_permissions: bytes = b''  # "Content access permissions (one bit for each content)"
        self.title_limits_list: List[_TitleLimit] = []  # List of play limits applied to the title.
        # v1 ticket data
        # TODO: Write in v1 ticket attributes here. This code can currently only handle v0 tickets, and will reject v1.

    def load(self, ticket: bytes) -> None:
        """
        Loads raw Ticket data and sets all attributes of the WAD object. This allows for manipulating an already
        existing Ticket.

        Parameters
        ----------
        ticket : bytes
            The data for the Ticket you wish to load.
        """
        with io.BytesIO(ticket) as ticket_data:
            # ====================================================================================
            # Parses each of the keys contained in the Ticket.
            # ====================================================================================
            # Signature type.
            ticket_data.seek(0x0)
            self.signature_type = ticket_data.read(4)
            # Signature data.
            ticket_data.seek(0x04)
            self.signature = ticket_data.read(256)
            # Signature issuer.
            ticket_data.seek(0x140)
            self.signature_issuer = str(ticket_data.read(64).replace(b'\x00', b'').decode())
            # ECDH data.
            ticket_data.seek(0x180)
            self.ecdh_data = ticket_data.read(60)
            # Ticket version.
            ticket_data.seek(0x1BC)
            self.ticket_version = int.from_bytes(ticket_data.read(1))
            if self.ticket_version == 1:
                raise ValueError("This appears to be a v1 ticket, which is not currently supported by libWiiPy. This "
                                 "feature is planned for a later release. Only v0 tickets are supported at this time.")
            # Title Key (Encrypted by a common key).
            ticket_data.seek(0x1BF)
            self.title_key_enc = ticket_data.read(16)
            # Ticket ID.
            ticket_data.seek(0x1D0)
            self.ticket_id = ticket_data.read(8)
            # Console ID.
            ticket_data.seek(0x1D8)
            self.console_id = int.from_bytes(ticket_data.read(4))
            # Title ID.
            ticket_data.seek(0x1DC)
            self.title_id = ticket_data.read(8)
            # Unknown data 1.
            ticket_data.seek(0x1E4)
            self.unknown1 = ticket_data.read(2)
            # Title version.
            ticket_data.seek(0x1E6)
            self.title_version = int.from_bytes(ticket_data.read(2))
            # Permitted titles mask.
            ticket_data.seek(0x1E8)
            self.permitted_titles = ticket_data.read(4)
            # Permit mask.
            ticket_data.seek(0x1EC)
            self.permit_mask = ticket_data.read(4)
            # Whether title export with a PRNG key is allowed.
            ticket_data.seek(0x1F0)
            self.title_export_allowed = int.from_bytes(ticket_data.read(1))
            # Common key index.
            ticket_data.seek(0x1F1)
            self.common_key_index = int.from_bytes(ticket_data.read(1))
            # Unknown data 2.
            ticket_data.seek(0x1F2)
            self.unknown2 = ticket_data.read(48)
            # Content access permissions.
            ticket_data.seek(0x222)
            self.content_access_permissions = ticket_data.read(64)
            # Content limits.
            ticket_data.seek(0x264)
            for limit in range(0, 8):
                limit_type = int.from_bytes(ticket_data.read(4))
                limit_value = int.from_bytes(ticket_data.read(4))
                self.title_limits_list.append(_TitleLimit(limit_type, limit_value))
        # Check certs to see if this is a retail or dev ticket. Treats unknown certs as being retail for now.
        if (self.signature_issuer.find("Root-CA00000002-XS00000006") != -1 or
                self.signature_issuer.find("Root-CA00000002-XS00000004") != -1):
            self.is_dev = True
        else:
            self.is_dev = False

    def dump(self) -> bytes:
        """
        Dumps the Ticket object back into bytes.

        Returns
        -------
        bytes
            The full Ticket file as bytes.
        """
        ticket_data = b''
        # Signature type.
        ticket_data += self.signature_type
        # Signature data.
        ticket_data += self.signature
        # Padding to 64 bytes.
        ticket_data += b'\x00' * 60
        # Signature issuer.
        signature_issuer = self.signature_issuer.encode()
        while len(signature_issuer) < 0x40:
            signature_issuer += b'\x00'
        ticket_data += signature_issuer
        # ECDH data.
        ticket_data += self.ecdh_data
        # Ticket version.
        ticket_data += int.to_bytes(self.ticket_version, 1)
        # Reserved (all \0x00).
        ticket_data += b'\x00\x00'
        # Title Key.
        ticket_data += self.title_key_enc
        # Unknown (write \0x00).
        ticket_data += b'\x00'
        # Ticket ID.
        ticket_data += self.ticket_id
        # Console ID.
        ticket_data += int.to_bytes(self.console_id, 4)
        # Title ID.
        ticket_data += self.title_id
        # Unknown data 1.
        ticket_data += self.unknown1
        # Title version.
        ticket_data += int.to_bytes(self.title_version, 2)
        # Permitted titles mask.
        ticket_data += self.permitted_titles
        # Permit mask.
        ticket_data += self.permit_mask
        # Title Export allowed.
        ticket_data += int.to_bytes(self.title_export_allowed, 1)
        # Common Key index.
        ticket_data += int.to_bytes(self.common_key_index, 1)
        # Unknown data 2.
        ticket_data += self.unknown2
        # Content access permissions.
        ticket_data += self.content_access_permissions
        # Padding (always \x00).
        ticket_data += b'\x00\x00'
        # Iterate over Title Limit objects, write them back into raw data, then add them to the Ticket.
        for title_limit in range(len(self.title_limits_list)):
            title_limit_data = b''
            # Write all fields from the title limit entry.
            title_limit_data += int.to_bytes(self.title_limits_list[title_limit].limit_type, 4)
            title_limit_data += int.to_bytes(self.title_limits_list[title_limit].maximum_usage, 4)
            # Write the entry to the ticket.
            ticket_data += title_limit_data
        return ticket_data

    def fakesign(self) -> None:
        """
        Fakesigns this Ticket for the trucha bug.

        This is done by brute-forcing a Ticket body hash starting with 00, causing it to pass signature verification on
        older IOS versions that incorrectly check the hash using strcmp() instead of memcmp(). The signature will also
        be erased and replaced with all NULL bytes.

        The hash is brute-forced by using the first two bytes of an unused section of the Ticket as a 16-bit integer,
        and incrementing that value by 1 until an appropriate hash is found.

        This modifies the Ticket object in place. You will need to call this method after any changes, and before
        dumping the Ticket object back into bytes.
        """
        # Clear the signature, so that the hash derived from it is guaranteed to always be
        # '0000000000000000000000000000000000000000'.
        self.signature = b'\x00' * 256
        current_int = 0
        test_hash = ''
        while test_hash[:2] != '00':
            current_int += 1
            # We're using the first 2 bytes of this unused region of the Ticket as a 16-bit integer, and incrementing
            # that to brute-force the hash we need.
            data_to_edit = self.unknown2
            data_to_edit = int.to_bytes(current_int, 2) + data_to_edit[2:]
            self.unknown2 = data_to_edit
            # Trim off the first 320 bytes, because we're only looking for the hash of the Ticket's body.
            # This is a try-except because an OverflowError will be thrown if the number being used to brute-force the
            # hash gets too big, as it is only a 16-bit integer. If that happens, then fakesigning has failed.
            try:
                test_hash = hashlib.sha1(self.dump()[320:]).hexdigest()
            except OverflowError:
                raise Exception("An error occurred during fakesigning. Ticket could not be fakesigned!")

    def get_is_fakesigned(self) -> bool:
        """
        Checks the Ticket object to see if it is currently fakesigned. For a description of fakesigning, refer to the
        fakesign() method.

        Returns
        -------
        bool:
            True if the Ticket is fakesigned, False otherwise.

        See Also
        --------
        libWiiPy.title.ticket.Ticket.fakesign()
        """
        if self.signature != b'\x00' * 256:
            return False
        test_hash = hashlib.sha1(self.dump()[320:]).hexdigest()
        if test_hash[:2] != '00':
            return False
        return True

    def get_title_id(self) -> str:
        """
        Gets the Title ID of the ticket's associated title.

        Returns
        -------
        str
            The Title ID of the title.
        """
        title_id_str = str(self.title_id.decode())
        return title_id_str

    def get_common_key_type(self) -> str:
        """
        Gets the name of the common key used to encrypt the Title Key contained in the ticket.

        Returns
        -------
        str
            The name of the common key required.

        See Also
        --------
        libWiiPy.title.commonkeys.get_common_key
        """
        match self.common_key_index:
            case 0:
                return "Common"
            case 1:
                return "Korean"
            case 2:
                return "vWii"
            case _:
                return "Unknown"

    def get_title_key(self) -> bytes:
        """
        Gets the decrypted title key contained in the ticket.

        Returns
        -------
        bytes
            The decrypted title key.
        """
        title_key = decrypt_title_key(self.title_key_enc, self.common_key_index, self.title_id, self.is_dev)
        return title_key

    def set_title_id(self, title_id) -> None:
        """
        Sets the Title ID property of the Ticket. Recommended over setting the property directly because of input
        validation.

        Parameters
        ----------
        title_id : str
            The new Title ID of the title.
        """
        if len(title_id) != 16:
            raise ValueError("Invalid Title ID! Title IDs must be 8 bytes long.")
        self.title_id = binascii.unhexlify(title_id.encode())

    def set_title_version(self, new_version: str | int) -> None:
        """
        Sets the version of the title in the Ticket. Recommended over setting the data directly because of input
        validation.

        Accepts either standard form (vX.X) as a string or decimal form (vXXX) as an integer.

        Parameters
        ----------
        new_version : str, int
            The new version of the title. See description for valid formats.
        """
        if type(new_version) is str:
            # Validate string input is in the correct format, then validate that the version isn't higher than v255.0.
            # If checks pass, convert to decimal form and set that as the title version.
            version_str_split = new_version.split(".")
            if len(version_str_split) != 2:
                raise ValueError("Title version is not valid! String version must be entered in format \"X.X\".")
            if int(version_str_split[0]) > 255 or int(version_str_split[1]) > 255:
                raise ValueError("Title version is not valid! String version number cannot exceed v255.255.")
            version_converted = title_ver_standard_to_dec(new_version, str(self.title_id.decode()))
            self.title_version = version_converted
        elif type(new_version) is int:
            # Validate that the version isn't higher than v65280. If the check passes, set that as the title version.
            if new_version > 65535:
                raise ValueError("Title version is not valid! Integer version number cannot exceed v65535.")
            self.title_version = new_version
        else:
            raise TypeError("Title version type is not valid! Type must be either integer or string.")



# "archive/u8.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki/U8_archive for details about the U8 archive format.

import binascii
import hashlib
import io
import os
import pathlib
from enum import IntEnum as _IntEnum
from dataclasses import dataclass as _dataclass
from typing import List, Tuple
from ..shared import _align_value, _pad_bytes


@_dataclass
class _U8Node:
    """
    A U8Node object that contains the data of a single node in a U8 file header. Each node keeps track of whether this
    node is for a file or directory, the offset of the name of the file/directory, the offset of the data for the file/
    directory, and the size of the data. Private class used by functions and methods in the U8 module.

    Attributes
    ----------
    type : int
        Whether this node refers to a file or a directory. Either 0x0000 for files, or 0x0100 for directories.
    name_offset : int
        The offset of the name of the file/directory this node refers to.
    data_offset : int
        The offset of the data for the file/directory this node refers to.
    size : int
        The size of the data for this node.
    """
    type: int
    name_offset: int
    data_offset: int
    size: int


class U8Archive:
    """
    A U8 object that allows for parsing and editing the contents of a U8 archive.

    Attributes
    ----------
    u8_node_list : List[_U8Node]
        A list of U8Node objects representing the nodes of the U8 archive.
    file_name_list : List[str]
        A list of the names of the files in the U8 archive.
    file_data_list : List[bytes]
        A list of file data for the files in the U8 archive; corresponds with file_name_list.
    header_size : int
        The size of the U8 archive header.
    data_offset : int
        The offset of the data region of the U8 archive.
    imet_header: IMETHeader
        The IMET header of the U8 archive, if one exists. Otherwise, an empty IMETHeader object.
    """
    def __init__(self):
        self.u8_magic = b''
        self.u8_node_list: List[_U8Node] = []  # All the nodes in the header of a U8 file.
        self.file_name_list: List[str] = []
        self.file_data_list: List[bytes] = []
        self.root_node_offset: int = 0
        self.header_size: int = 0
        self.data_offset: int = 0
        self.root_node: _U8Node = _U8Node(0, 0, 0, 0)
        self.imet_header: IMETHeader = IMETHeader()

    def load(self, u8_data: bytes) -> None:
        """
        Loads raw U8 data into a new U8 object. This allows for extracting the file and updating its contents.

        Parameters
        ----------
        u8_data : bytes
            The data for the U8 file to load.
        """
        with io.BytesIO(u8_data) as u8_data:
            # Read the first 4 bytes of the file to ensure that it's a U8 archive.
            u8_data.seek(0x0)
            self.u8_magic = u8_data.read(4)
            if self.u8_magic != b'\x55\xAA\x38\x2D':
                # Check for an IMET header, if the file doesn't start with the proper magic number. The header magic
                # may be at either 0x40 or 0x80 depending on whether this title has a build tag at the start or not.
                u8_data.seek(0x40)
                self.u8_magic = u8_data.read(4)
                if self.u8_magic == b'\x49\x4D\x45\x54':
                    # IMET with no build tag means the U8 archive should start at 0x600.
                    u8_data.seek(0x600)
                    self.u8_magic = u8_data.read(4)
                    if self.u8_magic != b'\x55\xAA\x38\x2D':
                        raise TypeError("This is not a valid U8 archive!")
                    # Parse the IMET header, then continue parsing the U8 archive.
                    u8_data.seek(0x0)
                    self.imet_header.load(u8_data.read(0x600))
                else:
                    # This check will pass if the IMET comes after a build tag.
                    u8_data.seek(0x80)
                    self.u8_magic = u8_data.read(4)
                    if self.u8_magic == b'\x49\x4D\x45\x54':
                        # IMET with a build tag means the U8 archive should start at 0x640.
                        u8_data.seek(0x640)
                        self.u8_magic = u8_data.read(4)
                        if self.u8_magic != b'\x55\xAA\x38\x2D':
                            raise TypeError("This is not a valid U8 archive!")
                        # Parse the IMET header, then continue parsing the U8 archive.
                        u8_data.seek(0x40)
                        self.imet_header.load(u8_data.read(0x600))
                    else:
                        raise TypeError("This is not a valid U8 archive!")
            # Offset of the root node, which will always be 0x20.
            self.root_node_offset = int.from_bytes(u8_data.read(4))
            # The size of the U8 header.
            self.header_size = int.from_bytes(u8_data.read(4))
            # The offset of the data, which is root_node_offset + header_size, aligned to 0x10.
            self.data_offset = int.from_bytes(u8_data.read(4))
            # Seek past 16 bytes of padding, then load the root node.
            u8_data.seek(u8_data.tell() + 16)
            root_node_type = int.from_bytes(u8_data.read(1))
            root_node_name_offset = int.from_bytes(u8_data.read(3))
            root_node_data_offset = int.from_bytes(u8_data.read(4))
            root_node_size = int.from_bytes(u8_data.read(4))
            self.root_node = _U8Node(root_node_type, root_node_name_offset, root_node_data_offset, root_node_size)
            # Seek back before the root node so that it gets read with all the rest.
            u8_data.seek(u8_data.tell() - 12)
            # Iterate over the number of nodes that the root node lists.
            for node in range(root_node_size):
                node_type = int.from_bytes(u8_data.read(1))
                node_name_offset = int.from_bytes(u8_data.read(3))
                node_data_offset = int.from_bytes(u8_data.read(4))
                node_size = int.from_bytes(u8_data.read(4))
                self.u8_node_list.append(_U8Node(node_type, node_name_offset, node_data_offset, node_size))
            # Iterate over all loaded nodes and create a list of file names and a list of file data.
            name_base_offset = u8_data.tell()
            for node in self.u8_node_list:
                u8_data.seek(name_base_offset + node.name_offset)
                name_bin = b''
                while name_bin[-1:] != b'\x00':
                    name_bin += u8_data.read(1)
                name_bin = name_bin[:-1]
                name = str(name_bin.decode())
                self.file_name_list.append(name)
                if node.type == 0:
                    u8_data.seek(node.data_offset)
                    self.file_data_list.append(u8_data.read(node.size))
                else:
                    self.file_data_list.append(b'')

    def dump(self) -> bytes:
        """
        Dumps the U8Archive object into the raw data of a U8 archive.

        Returns
        -------
        bytes
            The full U8 archive as bytes.
        """
        # This is 0 because the header size DOES NOT include the initial 32 bytes describing the file.
        header_size = 0
        # Add 12 bytes for each node, since that's how many bytes each one is made up of.
        for node in range(len(self.u8_node_list)):
            header_size += 12
        # Add the number of bytes used for each file/folder name in the string table.
        for file_name in self.file_name_list:
            header_size += len(file_name) + 1
        # The initial data offset is equal to the file header (32 bytes) + node data aligned to 64 bytes.
        data_offset = _align_value(header_size + 32, 64)
        # Adjust all nodes to place file data in the same order as the nodes. Why isn't it already like this?
        current_data_offset = data_offset
        current_name_offset = 0
        for node in range(len(self.u8_node_list)):
            if self.u8_node_list[node].type == 0:
                self.u8_node_list[node].data_offset = _align_value(current_data_offset, 32)
                current_data_offset += _align_value(self.u8_node_list[node].size, 32)
            # Calculate the name offsets, including the extra 1 for the NULL byte at the end of each name.
            self.u8_node_list[node].name_offset = current_name_offset
            current_name_offset += len(self.file_name_list[node]) + 1
        # Begin joining all the U8 archive data into bytes.
        u8_data = b''
        # Magic number.
        u8_data += b'\x55\xAA\x38\x2D'
        # Root node offset (this is always 0x20).
        u8_data += int.to_bytes(0x20, 4)
        # Size of the file header (excluding the first 32 bytes).
        u8_data += int.to_bytes(header_size, 4)
        # Offset of the beginning of the data region of the U8 archive.
        u8_data += int.to_bytes(data_offset, 4)
        # 16 bytes of zeroes.
        u8_data += (b'\x00' * 16)
        # Iterate over all the U8 nodes and dump them.
        for node in self.u8_node_list:
            u8_data += int.to_bytes(node.type, 1)
            u8_data += int.to_bytes(node.name_offset, 3)
            u8_data += int.to_bytes(node.data_offset, 4)
            u8_data += int.to_bytes(node.size, 4)
        # Iterate over all file names and dump them. All file names are suffixed by a \x00 byte.
        for file_name in self.file_name_list:
            u8_data += str.encode(file_name) + b'\x00'
        # Apply the extra padding we calculated earlier by padding to where the data offset begins.
        u8_data = _pad_bytes(u8_data, 64)
        # Iterate all file data and dump it.
        for file in self.file_data_list:
            u8_data += _pad_bytes(file, 32)
        # Return the U8 archive.
        return u8_data


def extract_u8(u8_data, output_folder) -> None:
    """
    Extracts the provided U8 archive file data into the provided output folder path. Note that the folder must not
    already exist to ensure that the output can correctly represent the file structure of the original U8 archive.

    Parameters
    ----------
    u8_data : bytes
        The data for the U8 file to extract.
    output_folder : str
        The path to a new folder to extract the archive to.
    """
    output_folder = pathlib.Path(output_folder)
    # Check if the path already exists, and if it does, ensure that it is both a directory and empty.
    if output_folder.exists():
        if output_folder.is_dir() and next(os.scandir(output_folder), None):
            raise ValueError("Output folder is not empty!")
        elif output_folder.is_file():
            raise ValueError("A file already exists with the provided name!")
    else:
        os.mkdir(output_folder)
    # Create a new U8Archive object and load the provided U8 file data into it.
    u8_archive = U8Archive()
    u8_archive.load(u8_data)
    # This variable stores the path of the directory we're currently processing.
    current_dir = output_folder
    # This variable stores the order of directory nodes leading to the current working directory, to make sure that
    # things get where they belong.
    parent_dirs = [0]
    for node in range(len(u8_archive.u8_node_list)):
        # Code for a directory node (excluding the root node since that already exists).
        if u8_archive.u8_node_list[node].type == 1 and u8_archive.u8_node_list[node].name_offset != 0:
            if u8_archive.u8_node_list[node].data_offset == parent_dirs[-1]:
                current_dir = current_dir.joinpath(u8_archive.file_name_list[node])
                current_dir.mkdir(exist_ok=True)
                parent_dirs.append(node)
            else:
                # Go up until we're back at the correct level.
                while u8_archive.u8_node_list[node].data_offset != parent_dirs[-1]:
                    parent_dirs.pop()
                parent_dirs.append(node)
                current_dir = output_folder
                # Rebuild current working directory, and make sure all directories in the path exist.
                for directory in parent_dirs:
                    current_dir = current_dir.joinpath(u8_archive.file_name_list[directory])
                    current_dir.mkdir(exist_ok=True)
        # Code for a file node.
        elif u8_archive.u8_node_list[node].type == 0:
            open(current_dir.joinpath(u8_archive.file_name_list[node]), "wb").write(u8_archive.file_data_list[node])
        # Handle an invalid node type.
        elif u8_archive.u8_node_list[node].type != 0 and u8_archive.u8_node_list[node].type != 1:
            raise ValueError(f"A node with an invalid type ({str(u8_archive.u8_node_list[node].type)}) was found!")


def _pack_u8_dir(u8_archive: U8Archive, current_path, node_count, parent_node):
    # First, get the list of everything in current path.
    root_list = os.listdir(current_path)
    file_list = []
    dir_list = []
    # Create separate lists of the files and directories in the current directory so that we can handle the files first.
    # noinspection PyTypeChecker
    root_list.sort(key=str.lower)
    for path in root_list:
        if os.path.isfile(current_path.joinpath(path)):
            file_list.append(path)
        elif os.path.isdir(current_path.joinpath(path)):
            dir_list.append(path)
    # noinspection PyTypeChecker
    file_list.sort(key=str.lower)
    # noinspection PyTypeChecker
    dir_list.sort(key=str.lower)
    # For files, read their data into the file data list, add their name into the file name list, then calculate the
    # offset for their file name and create a new U8Node() for them. -1 values are temporary and are set during dumping.
    for file in file_list:
        node_count += 1
        u8_archive.file_name_list.append(file)
        u8_archive.file_data_list.append(open(current_path.joinpath(file), "rb").read())
        u8_archive.u8_node_list.append(_U8Node(0, -1, -1, len(u8_archive.file_data_list[-1])))
    # For directories, add their name to the file name list, add empty data to the file data list (since they obviously
    # wouldn't have any), find the total number of files and directories inside the directory to calculate the final
    # node included in it, then recursively call this function again on that directory to process it.
    for directory in dir_list:
        node_count += 1
        u8_archive.file_name_list.append(directory)
        u8_archive.file_data_list.append(b'')
        max_node = node_count + sum(1 for _ in current_path.joinpath(directory).rglob('*'))
        u8_archive.u8_node_list.append(_U8Node(1, -1, parent_node, max_node))
        u8_archive, node_count = _pack_u8_dir(u8_archive, current_path.joinpath(directory), node_count,
                                              u8_archive.u8_node_list.index(u8_archive.u8_node_list[-1]))
    # Return the U8Archive object, the current node we're on, and the current name offset.
    return u8_archive, node_count


def pack_u8(input_path, generate_imet=False, imet_titles:List[str]=None) -> bytes:
    """
    Packs the provided file or folder into a new U8 archive, and returns the raw file data for it.

    To generate an IMET header for this U8 archive, the archive must contain the required banner files "icon.bin",
    "banner.bin", and "sound.bin", because the sizes of these files are stored in the header.

    Parameters
    ----------
    input_path
        The path to the input file or folder.
    generate_imet : bool, optional
        Whether an IMET header should be generated for this U8 archive or not. IMET headers are only used for channel
        banners (00000000.app), and required banner files must exist to generate this header. Defaults to False.
    imet_titles : List[str], optional
        A list of the channel title in different languages for the IMET header. If only one item is provided, that
        item will be used for all entries in the header. Defaults to None, and is only used when generate_imet is True.

    Returns
    -------
    u8_archive : bytes
        The data for the packed U8 archive.
    """
    input_path = pathlib.Path(input_path)
    if input_path.is_dir():
        # Append empty entries at the start for the root node, and then create the root U8Node() object, using rglob()
        # to read the total count of files and directories that will be packed so that we can add the total node count.
        u8_archive = U8Archive()
        u8_archive.file_name_list.append("")
        u8_archive.file_data_list.append(b'')
        u8_archive.u8_node_list.append(_U8Node(1, 0, 0, sum(1 for _ in input_path.rglob('*')) + 1))
        # Call the private function _pack_u8_dir() on the root note, which will recursively call itself to pack every
        # subdirectory and file. Discard node_count and name_offset since we don't care about them here, as they're
        # really only necessary for the directory recursion.
        u8_archive, _ = _pack_u8_dir(u8_archive, input_path, node_count=1, parent_node=0)
        if generate_imet:
            print("gen imet")
        return u8_archive.dump()
    elif input_path.is_file():
        raise ValueError("This does not appear to be a directory.")
    else:
        raise FileNotFoundError(f"Input directory: \"{input_path}\" does not exist!")


class IMETHeader:
    """
    An IMETHeader object that allows for parsing, editing, and generating an IMET header. These headers precede the
    data of a channel banner (00000000.app), and are used to store metadata about the banner and verify its data.

    An IMET header is always 1,536 (0x600) bytes long.

    Attributes
    ----------
    magic : str
        Magic number for the header, should be "IMD5".
    header_size : int
        Length of the M
    imet_version : int
        Version of the IMET header. Normally always 3.
    sizes : List[int]
        The file sizes of icon.bin, banner.bin, and sound.bin.
    flag1 : int
        Unknown.
    channel_names : List[str]
        The name of the channel this header is for in Japanese, English, German, French, Spanish, Italian, Dutch,
        Simplified Chinese, Traditional Chinese, and Korean, in that order.
    md5_hash : bytes
        MD5 sum of the entire header, with this field being all zeros during the hashing.
    """
    def __init__(self):
        self.magic: str = ""  # Should always be "IMET"
        self.header_size: int = 0  # Always 1536? I assumed this would mean something, but it's just the header length.
        self.imet_version: int = 0  # Always 3?
        self.sizes: List[int] = []  # Should only have 3 items
        self.flag1: int = 0  # Unknown
        self.channel_names: List[str] = []  # Should have 10 items
        self.md5_hash: bytes = b''

    class LocalizedTitles(_IntEnum):
        TITLE_JAPANESE = 0
        TITLE_ENGLISH = 1
        TITLE_GERMAN = 2
        TITLE_FRENCH = 3
        TITLE_SPANISH = 4
        TITLE_ITALIAN = 5
        TITLE_DUTCH = 6
        TITLE_CHINESE_SIMPLIFIED = 7
        TITLE_CHINESE_TRADITIONAL = 8
        TITLE_KOREAN = 9

    def load(self, imet_data: bytes) -> None:
        """
        Loads the raw data of an IMET header.

        Parameters
        ----------
        imet_data : bytes
            The data for the IMET header to load.
        """
        with io.BytesIO(imet_data) as data:
            data.seek(0x40)
            self.magic = str(data.read(4).decode())
            self.header_size = int.from_bytes(data.read(4))
            self.imet_version = int.from_bytes(data.read(4))
            self.sizes = []
            for _ in range(0, 3):
                self.sizes.append(int.from_bytes(data.read(4)))
            self.flag1 = int.from_bytes(data.read(4))
            self.channel_names = []
            for _ in range(0, 10):
                # Read the translated channel name from the header, then drop all trailing null bytes. The encoding
                # used here is UTF-16 Big Endian.
                new_channel_name = data.read(84)
                self.channel_names.append(str(new_channel_name.decode('utf-16-be')).replace('\x00', ''))
            data.seek(data.tell() + 588)
            self.md5_hash = binascii.hexlify(data.read(16))

    def dump(self) -> bytes:
        """
        Dump the IMETHeader back into raw bytes.

        Returns
        -------
        bytes
            The IMET header as bytes.
        """
        imet_data = b''
        # 64 bytes of padding.
        imet_data += b'\x00' * 64
        # "IMET" magic number.
        imet_data += str.encode("IMET")
        # IMET header size. TODO: check if this is actually always 1536
        imet_data += int.to_bytes(1536, 4)
        # IMET header version.
        imet_data += int.to_bytes(self.imet_version, 4)
        # Banner component sizes.
        for size in self.sizes:
            imet_data += int.to_bytes(size, 4)
        # flag1.
        imet_data += int.to_bytes(self.flag1, 4)
        # Channel names.
        for channel_name in self.channel_names:
            new_channel_name = channel_name.encode('utf-16-be')
            while len(new_channel_name) < 84:
                new_channel_name += b'\x00'
            imet_data += new_channel_name
        # 588 (WHY??) bytes of padding.
        imet_data += b'\x00' * 588
        # MD5 hash. To calculate the real one, we need to write all zeros to it first, then hash the entire header with
        # the zero hash. After that we'll replace this hash with the calculated one.
        imet_data += b'\x00' * 16
        imet_hash = hashlib.md5(imet_data)
        imet_data = imet_data[:-16] + imet_hash.digest()
        return imet_data

    def create(self, sizes: List[int], channel_names: Tuple[int, str] | List[Tuple[int, str]]) -> None:
        """
        Create a new IMET header, specifying the sizes of the banner components and one or more localized channel names.

        Parameters
        ----------
        sizes : List[int]
            The size in bytes of icon.bin, banner.bin, and sound.bin, in that order.
        channel_names : Tuple(int, str), List[Tuple[int, str]]
            A pair or list of pairs of the target language and channel name for that language. Target languages are
            defined in LocalizedTitles.

        See Also
        --------
        libWiiPy.archive.u8.IMETHeader.LocalizedTitles
        """
        # Begin by setting the constant values before we parse the input.
        self.magic = "IMET"
        self.header_size = 1536
        self.imet_version = 3
        self.flag1 = 0  # Still not really sure about this one.
        # Validate the number of entries, then set the provided file sizes.
        if len(sizes) != 3:
            raise ValueError("You must supply 3 file sizes to generate an IMET header!")
        self.sizes = sizes
        # Now we can parse the channel names. This functions the same as setting them later, so just calling
        # set_channel_names() is the most practical.
        self.channel_names = ["" for _ in range(0, 10)]
        self.set_channel_names(channel_names)

    def get_channel_names(self, target_languages: int | List[int]) -> str | List[str]:
        """
        Get one or more channel names from the IMET header based on the specified languages.

        Parameters
        ----------
        target_languages : int, List[int, str]
            One or more target languages. Target languages are defined in LocalizedTitles.

        Returns
        -------
        str, List[str]
            The channel name for the specified language, or a list of channel names in the same order as the specified
            languages.

        See Also
        --------
        libWiiPy.archive.u8.IMETHeader.LocalizedTitles
        """
        # Flatten single instance of LocalizedTitles being passed to a proper int.
        if isinstance(target_languages, self.LocalizedTitles):
            target_languages = int(target_languages)
        # If only one channel name was requested.
        if type(target_languages) == int:
            if target_languages not in self.LocalizedTitles:
                raise ValueError(f"The specified language is not valid!")
            return self.channel_names[target_languages]
        # If multiple channel names were requested.
        else:
            channel_names = []
            for lang in target_languages:
                if lang not in self.LocalizedTitles:
                    raise ValueError(f"The specified language at index {target_languages.index(lang)} is not valid!")
                channel_names.append(self.channel_names[lang])
            return channel_names

    def set_channel_names(self, channel_names: Tuple[int, str] | List[Tuple[int, str]]) -> None:
        """
        Specify one or more new channel names to set in the IMET header.

        Parameters
        ----------
        channel_names : Tuple(int, str), List[Tuple[int, str]]
            A pair or list of pairs of the target language and channel name for that language. Target languages are
            defined in LocalizedTitles.

        See Also
        --------
        libWiiPy.archive.u8.IMETHeader.LocalizedTitles
        """
        # If only one channel name was provided.
        if type(channel_names) == tuple:
            if channel_names[0] not in self.LocalizedTitles:
                raise ValueError(f"The target language \"{channel_names[0]}\" is not valid!")
            if len(channel_names[1]) > 42:
                raise ValueError(f"The channel name \"{channel_names[1]}\" is too long! Channel names cannot exceed "
                                 f"42 characters!")
            self.channel_names[channel_names[0]] = channel_names[1]
        # If a list of channel names was provided.
        else:
            for name in channel_names:
                if name[0] not in self.LocalizedTitles:
                    raise ValueError(f"The target language \"{name[0]}\" for the name at index {channel_names.index(name)} "
                                     f"is not valid!")
                if len(name[1]) > 42:
                    raise ValueError(f"The channel name \"{name[1]}\" at index {channel_names.index(name)} is too long! "
                                     f"Channel names cannot exceed 42 characters!")
                self.channel_names[name[0]] = name[1]



# "title/crypto.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy

import binascii
import struct
from .commonkeys import get_common_key
from Crypto.Cipher import AES as _AES


def _convert_tid_to_iv(title_id: str | bytes) -> bytes:
    # Converts a Title ID in various formats into the format required to act as an IV. Private function used by other
    # crypto functions.
    title_key_iv = b''
    if type(title_id) is bytes:
        # This catches the format b'0000000100000002'
        if len(title_id) == 16:
            title_key_iv = binascii.unhexlify(title_id)
        # This catches the format b'\x00\x00\x00\x01\x00\x00\x00\x02'
        elif len(title_id) == 8:
            title_key_iv = title_id
        # If it isn't one of those lengths, it cannot possibly be valid, so reject it.
        else:
            raise ValueError("Title ID is not valid!")
    # Allow for a string like "0000000100000002"
    elif type(title_id) is str:
        title_key_iv = binascii.unhexlify(title_id)
    # If the Title ID isn't bytes or a string, it isn't valid and is rejected.
    else:
        raise TypeError("Title ID type is not valid! It must be either type str or bytes.")
    return title_key_iv


def decrypt_title_key(title_key_enc: bytes, common_key_index: int, title_id: bytes | str, dev=False) -> bytes:
    """
    Gets the decrypted version of the encrypted Title Key provided.

    Requires the index of the common key to use, and the Title ID of the title that the Title Key is for.

    Parameters
    ----------
    title_key_enc : bytes
        The encrypted Title Key.
    common_key_index : int
        The index of the common key used to encrypt the Title Key.
    title_id : bytes, str
        The Title ID of the title that the key is for.
    dev : bool
        Whether the Title Key is encrypted with the development key or not.

    Returns
    -------
    bytes
        The decrypted Title Key.
    """
    # Load the correct common key for the title.
    common_key = get_common_key(common_key_index, dev)
    # Convert the IV into the correct format based on the type provided.
    title_key_iv = _convert_tid_to_iv(title_id)
    # The IV will always be in the same format by this point, so add the last 8 bytes.
    title_key_iv = title_key_iv + (b'\x00' * 8)
    # Create a new AES object with the values provided.
    aes = _AES.new(common_key, _AES.MODE_CBC, title_key_iv)
    # Decrypt the Title Key using the AES object.
    title_key = aes.decrypt(title_key_enc)
    return title_key


def encrypt_title_key(title_key_dec: bytes, common_key_index: int, title_id: bytes | str, dev=False) -> bytes:
    """
    Encrypts the provided Title Key with the selected common key.

    Requires the index of the common key to use, and the Title ID of the title that the Title Key is for.

    Parameters
    ----------
    title_key_dec : bytes
        The decrypted Title Key.
    common_key_index : int
        The index of the common key used to encrypt the Title Key.
    title_id : bytes, str
        The Title ID of the title that the key is for.
    dev : bool
        Whether the Title Key is encrypted with the development key or not.

    Returns
    -------
    bytes
        An encrypted Title Key.
    """
    # Load the correct common key for the title.
    common_key = get_common_key(common_key_index, dev)
    # Convert the IV into the correct format based on the type provided.
    title_key_iv = _convert_tid_to_iv(title_id)
    # The IV will always be in the same format by this point, so add the last 8 bytes.
    title_key_iv = title_key_iv + (b'\x00' * 8)
    # Create a new AES object with the values provided.
    aes = _AES.new(common_key, _AES.MODE_CBC, title_key_iv)
    # Encrypt Title Key using the AES object.
    title_key = aes.encrypt(title_key_dec)
    return title_key


def decrypt_content(content_enc, title_key, content_index, content_length) -> bytes:
    """
    Gets the decrypted version of the encrypted content.

    This requires the index of the content to decrypt as it is used as the IV, as well as the content length to adjust
    padding as necessary.

    Parameters
    ----------
    content_enc : bytes
        The encrypted content.
    title_key : bytes
        The Title Key for the title the content is from.
    content_index : int
        The index in the TMD's content record of the content being decrypted.
    content_length : int
        The length in the TMD's content record of the content being decrypted.

    Returns
    -------
    bytes
        The decrypted content.
    """
    # Generate the IV from the Content Index of the content to be decrypted.
    content_index_bin = struct.pack('>H', content_index)
    while len(content_index_bin) < 16:
        content_index_bin += b'\x00'
    # Align content to 16 bytes to ensure that it works with AES encryption.
    if (len(content_enc) % 16) != 0:
        content_enc = content_enc + (b'\x00' * (16 - (len(content_enc) % 16)))
    # Create a new AES object with the values provided, with the content's unique ID as the IV.
    aes = _AES.new(title_key, _AES.MODE_CBC, content_index_bin)
    # Decrypt the content using the AES object.
    content_dec = aes.decrypt(content_enc)
    # Trim additional bytes that may have been added so the content is the correct size.
    content_dec = content_dec[:content_length]
    return content_dec


def encrypt_content(content_dec, title_key, content_index) -> bytes:
    """
    Gets the encrypted version of the decrypted content.

    This requires the index of the content to encrypt as it is used as the IV, as well as the content length to adjust
    padding as necessary.

    Parameters
    ----------
    content_dec : bytes
        The decrypted content.
    title_key : bytes
        The Title Key for the title the content is from.
    content_index : int
        The index in the TMD's content record of the content being decrypted.

    Returns
    -------
    bytes
        The encrypted content.
    """
    # Generate the IV from the Content Index of the content to be decrypted.
    content_index_bin = struct.pack('>H', content_index)
    while len(content_index_bin) < 16:
        content_index_bin += b'\x00'
    # Calculate the intended size of the encrypted content.
    enc_size = len(content_dec) + (16 - (len(content_dec) % 16))
    # Align content to 16 bytes to ensure that it works with AES encryption.
    if (len(content_dec) % 16) != 0:
        content_dec = content_dec + (b'\x00' * (16 - (len(content_dec) % 16)))
    # Create a new AES object with the values provided, with the content's unique ID as the IV.
    aes = _AES.new(title_key, _AES.MODE_CBC, content_index_bin)
    # Encrypt the content using the AES object.
    content_enc = aes.encrypt(content_dec)
    # Trim down the encrypted content.
    content_enc = content_enc[:enc_size]
    return content_enc



# "title/banner.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki/Opening.bnr for details about the Wii's banner format

from dataclasses import dataclass as _dataclass


@_dataclass
class IMD5Header:
    """
    An IMD5Header object that contains the properties of an IMD5 header. These headers precede the data of banner.bin
    and icon.bin inside the banner (00000000.app) of a channel, and are used to verify the data of those files.

    An IMD5 header is always 32 bytes long.

    Attributes
    ----------
    magic : str
        Magic number for the header, should be "IMD5".
    file_size : int
        The size of the file this header precedes.
    zeros : int
        8 bytes of zero padding.
    md5_hash : bytes
        The MD5 hash of the file this header precedes.
    """
    magic: str  # Should always be "IMD5"
    file_size: int
    zeros: int
    md5_hash: bytes



# "title/cert.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki/Certificate_chain for details about the Wii's certificate chain

import io
from enum import IntEnum as _IntEnum
from ..shared import _align_value, _pad_bytes
from .ticket import Ticket
from .tmd import TMD
from Crypto.Hash import SHA1
from Crypto.PublicKey import RSA
from Crypto.Signature import pkcs1_15


class CertificateType(_IntEnum):
    RSA_4096 = 0x00010000
    RSA_2048 = 0x00010001
    ECC = 0x00010002


class CertificateSignatureLength(_IntEnum):
    RSA_4096 = 0x200
    RSA_2048 = 0x100
    ECC = 0x3C


class CertificateKeyType(_IntEnum):
    RSA_4096 = 0x00000000
    RSA_2048 = 0x00000001
    ECC = 0x00000002


class CertificateKeyLength(_IntEnum):
    RSA_4096 = 0x200
    RSA_2048 = 0x100
    ECC = 0x3C


class Certificate:
    """
    A Certificate object used to parse a certificate used for the Wii's content verification.

    Attributes
    ----------
    type: CertificateType
        The type of the certificate, either RSA-2048, RSA-4096, or ECC.
    signature: bytes
        The signature data of the certificate.
    issuer: str
        The certificate that issued this certificate.
    pub_key_type: CertificateKeyType
        The type of public key contained in the certificate, either RSA-2048, RSA-4096, or ECC.
    child_name: str
        The name of this certificate.
    pub_key_id: int
        The ID of this certificate's public key.
    pub_key_modulus: int
        The modulus of this certificate's public key. Combined with the exponent to get the full key.
    pub_key_exponent: int
        The exponent of this certificate's public key. Combined with the modulus to get the full key.
    """
    def __init__(self):
        self.type: CertificateType | None = None
        self.signature: bytes = b''
        self.issuer: str = ""
        self.pub_key_type: CertificateKeyType | None = None
        self.child_name: str = ""
        self.pub_key_id: int = 0
        self.pub_key_modulus: int = 0
        self.pub_key_exponent: int = 0

    def load(self, cert: bytes) -> None:
        """
        Loads certificate data into the Certificate object, allowing you to parse the certificate.

        Parameters
        ----------
        cert: bytes
            The data for the certificate to load.
        """
        with io.BytesIO(cert) as cert_data:
            # Read the first 4 bytes of the cert to get the certificate's type.
            try:
                self.type = CertificateType.from_bytes(cert_data.read(0x4))
            except ValueError:
                raise ValueError("Invalid Certificate Type!")
            cert_length = CertificateSignatureLength[self.type.name]
            self.signature = cert_data.read(cert_length.value)
            cert_data.seek(0x40 + cert_length.value)
            self.issuer = str(cert_data.read(0x40).replace(b'\x00', b'').decode())
            try:
                cert_data.seek(0x80 + cert_length.value)
                self.pub_key_type = CertificateKeyType.from_bytes(cert_data.read(0x4))
            except ValueError:
                raise ValueError("Invalid Certificate Key type!")
            cert_data.seek(0x84 + cert_length.value)
            self.child_name = str(cert_data.read(0x40).replace(b'\x00', b'').decode())
            cert_data.seek(0xC4 + cert_length.value)
            self.pub_key_id = int.from_bytes(cert_data.read(0x4))
            key_length = CertificateKeyLength[self.pub_key_type.name]
            cert_data.seek(0xC8 + cert_length.value)
            self.pub_key_modulus = int.from_bytes(cert_data.read(key_length.value))
            if self.pub_key_type == CertificateKeyType.RSA_4096 or self.pub_key_type == CertificateKeyType.RSA_2048:
                self.pub_key_exponent = int.from_bytes(cert_data.read(0x4))

    def dump(self) -> bytes:
        """
        Dump the certificate object back into bytes.

        Returns
        -------
        bytes:
            The certificate file as bytes.
        """
        cert_data = b''
        cert_data += int.to_bytes(self.type.value, 4)
        cert_data += self.signature
        cert_data = _pad_bytes(cert_data)
        # Pad out the issuer name with null bytes.
        issuer = self.issuer.encode()
        while len(issuer) < 0x40:
            issuer += b'\x00'
        cert_data += issuer
        cert_data += int.to_bytes(self.pub_key_type.value, 4)
        # Pad out the child cert name with null bytes
        child_name = self.child_name.encode()
        while len(child_name) < 0x40:
            child_name += b'\x00'
        cert_data += child_name
        cert_data += int.to_bytes(self.pub_key_id, 4)
        cert_data += int.to_bytes(self.pub_key_modulus, CertificateKeyLength[self.pub_key_type.name])
        if self.pub_key_type == CertificateKeyType.RSA_4096 or self.pub_key_type == CertificateKeyType.RSA_2048:
            cert_data += int.to_bytes(self.pub_key_exponent, 4)
        # Pad out the certificate data to a multiple of 64.
        cert_data = _pad_bytes(cert_data)
        return cert_data


class CertificateChain:
    """
    A CertificateChain object used to parse the chain of certificates stored in a WAD that are used for the Wii's
    content verification. The certificate chain is the format that the certificates are stored in as part of every WAD.

    Attributes
    ----------
    ca_cert: Certificate
        The CA certificate from the chain.
    tmd_cert: Certificate
        The CP (TMD) certificate from the chain.
    ticket_cert: Certificate
        The XS (Ticket) certificate from the chain.
    """
    def __init__(self):
        self.ca_cert: Certificate = Certificate()
        self.tmd_cert: Certificate = Certificate()
        self.ticket_cert: Certificate = Certificate()

    def load(self, cert_chain: bytes) -> None:
        """
        Loads certificate chain data into the CertificateChain object, allowing you to parse the individual
        certificates stored in the chain.

        Parameters
        ----------
        cert_chain: bytes
            The data for the certificate chain to load.
        """
        with (io.BytesIO(cert_chain) as cert_chain_data):
            # Read the two fields that denote different length sections of the certificate, so that we know how long
            # this certificate is in total.
            offset = 0x0
            for _ in range(3):
                cert_chain_data.seek(offset)
                cert_type = CertificateType.from_bytes(cert_chain_data.read(0x4))
                cert_chain_data.seek(offset + 0x80 + CertificateSignatureLength[cert_type.name].value)
                key_type = CertificateKeyType.from_bytes(cert_chain_data.read(0x4))
                cert_size = _align_value(0xC8 + CertificateSignatureLength[cert_type.name].value +
                                               CertificateKeyLength[key_type.name].value)
                cert_chain_data.seek(offset + 0x0)
                cert = Certificate()
                cert.load(cert_chain_data.read(cert_size))
                if cert.issuer == "Root":
                    self.ca_cert = cert
                elif cert.issuer.find("Root-CA") != -1:
                    if cert.child_name.find("CP") != -1:
                        self.tmd_cert = cert
                    elif cert.child_name.find("XS") != -1:
                        self.ticket_cert = cert
                    else:
                        raise ValueError("Unknown certificate in chain!")
                else:
                    raise ValueError("Unknown certificate in chain!")
                offset += cert_size

    def dump(self) -> bytes:
        """
        Dumps the full certificate chain back into bytes. This chain will always be formatted with the CA cert first,
        followed by the CP (TMD) cert, then finally the XS (Ticket) cert.

        Returns
        -------
        bytes
            The full certificate chain as bytes.
        """
        cert_chain_data = b''
        cert_chain_data += self.ca_cert.dump()
        cert_chain_data += self.tmd_cert.dump()
        cert_chain_data += self.ticket_cert.dump()
        return cert_chain_data


def verify_ca_cert(ca_cert: Certificate) -> bool:
    """
    Verify a Wii CA certificate using the root public key. The retail or development root key will be automatically
    selected based off of the name of the CA certificate provided.

    Parameters
    ----------
    ca_cert: Certificate
        The CA certificate to verify.

    Returns
    -------
    bool
        Whether the certificate is valid or not.
    """
    if ca_cert.issuer != "Root" or ca_cert.child_name.find("CA") == -1:
        raise ValueError("The provided certificate is not a CA certificate!")
    if ca_cert.child_name == "CA00000001":
        root_key_modulus = \
            (b'\xf8$lX\xba\xe7P\x03\x01\xfb\xb7\xc2\xeb\xe0\x01\x05q\xda\x92#x\xf0QN\xc0\x03\x1d\xd0\xd2\x1e\xd3\xd0~'
             b'\xfc\x85 i\xb5\xde\x9b\xb9Q\xa8\xbc\x90\xa2D\x92m7\x92\x95\xae\x946\xaa\xa6\xa3\x02Q\x0c{\x1d\xed\xd5'
             b'\xfb \x86\x9d\x7f0\x16\xf6\xbee\xd3\x83\xa1m\xb32\x1b\x955\x18\x90\xb1p\x02\x93~\xe1\x93\xf5~\x99\xa2GN'
             b'\x9d8$\xc7\xae\xe3\x85A\xf5g\xe7Q\x8cz\x0e8\xe7\xeb\xafA\x19\x1b\xcf\xf1{B\xa6\xb4\xed\xe6\xce\x8d\xe71'
             b'\x8f\x7fR\x04\xb3\x99\x0e"gE\xaf\xd4\x85\xb2D\x93\x00\x8b\x08\xc7\xf6\xb7\xe5k\x02\xb3\xe8\xfe\x0c\x9d'
             b'\x85\x9c\xb8\xb6\x82#\xb8\xab\'\xee_e8\x07\x8b-\xb9\x1e*\x15>\x85\x81\x80r\xa2;m\xd92\x81\x05Oo\xb0\xf6'
             b'\xf5\xad(>\xca\x0bz\xf3TU\xe0=\xa7\xb6\x83&\xf3\xec\x83J\xf3\x14\x04\x8a\xc6\xdf \xd2\x85\x08g<\xabb\xa2'
             b'\xc7\xbc\x13\x1aS>\x0bf\x80k\x1c0fK7#1\xbd\xc4\xb0\xca\xd8\xd1\x1e\xe7\xbb\xd9(UH\xaa\xec\x1ff\xe8!\xb3'
             b'\xc8\xa0Gi\x00\xc5\xe6\x88\xe8\x0c\xce<a\xd6\x9c\xbb\xa17\xc6`Ozr\xdd\x8c{>=Q)\r\xaajY{\x08\x1f\x9d63'
             b'\xa3Fz5a\t\xac\xa7\xdd}./\xb2\xc1\xae\xb8\xe2\x0fH\x92\xd8\xb9\xf8\xb4oN<\x11\xf4\xf4}\x8bu}\xfe\xfe\xa3'
             b'\x89\x9c3Y\\^\xfd\xeb\xcb\xab\xe8A>:\x9a\x80<i5n\xb2\xb2\xad\\\xc4\xc8XE^\xf5\xf7\xb3\x06D\xb4|d\x06\x8c'
             b'\xdf\x80\x9fv\x02Z-\xb4F\xe0=|\xf6/4\xe7\x02E{\x02\xa4\xcf]\x9d\xd5<\xa5:|\xa6)x\x8cg\xca\x08\xbf\xec'
             b'\xcaC\xa9W\xad\x16\xc9N\x1c\xd8u\xca\x10}\xce~\x01\x18\xf0\xdfk\xfe\xe5\x1d\xdb\xd9\x91\xc2n`\xcdHX\xaa'
             b'Y,\x82\x00u\xf2\x9fRl\x91|o\xe5@>\xa7\xd4\xa5\x0c\xec;s\x84\xde\x88n\x82\xd2\xebMNB\xb5\xf2\xb1I\xa8\x1e'
             b'\xa7\xceqD\xdc)\x94\xcf\xc4N\x1f\x91\xcb\xd4\x95')
    elif ca_cert.child_name == "CA00000002":
        root_key_modulus = \
            (b'\x00\xd0\x1f\xe1\x00\xd45V\xb2KV\xda\xe9q\xb5\xa5\xd3\x84\xb90\x03\xbe\x1b\xbf(\xa20[\x06\x06EF}[\x02Q'
             b'\xd2V\x1a\'O\x9e\x9f\x9c\xecdaP\xab=*\xe36hf\xac\xa4\xba\xe8\x1a\xe3\xd7\x9a\xa6\xb0J\x8b\xcb\xa7\xe6'
             b'\xfbd\x89E\xeb\xdf\xdb\x85\xba\t\x1f\xd7\xd1\x14\xb5\xa3\xa7\x80\xe3\xa2.n\xcd\x87\xb5\xa4\xc6\xf9\x10'
             b'\xe4\x03"\x08\x81K\x0c\xee\xa1\xa1}\xf79i_a~\xf65(\xdb\x94\x967\xa0V\x03\x7f{2A8\x95\xc0\xa8\xf1\x98.'
             b'\x15e\xe3\x8e\xed\xc2.Y\x0e\xe2g{\x86\t\xf4\x8c.0?\xbc@\\\xac\x18\x04/\x82 \x84\xe4\x93h\x03\xda\x7fA4'
             b'\x92HV+\x8e\xe1/x\xf8\x03$c0\xbc{\xe7\xeerJ\xf4X\xa4r\xe7\xabF\xa1\xa7\xc1\x0c/\x18\xfa\x07\xc3\xdd\xd8'
             b'\x98\x06\xa1\x1c\x9c\xc10\xb2G\xa3<\x8dG\xdeg\xf2\x9eUw\xb1\x1cCI=[\xbav4\xa7\xe4\xe7\x151\xb7\xdfY\x81'
             b'\xfe$\xa1\x14UL\xbd\x8f\x00\\\xe1\xdb5\x08\\\xcf\xc7x\x06\xb6\xde%@h\xa2l\xb5I-E\x80C\x8f\xe1\xe5\xa9'
             b'\xedu\xc5\xedE\x1d\xcex\x949\xcc\xc3\xba(\xa21*\x1b\x87\x19\xef\x0fs\xb7\x13\x95\x0c\x02Y\x1atb\xa6\x07'
             b'\xf3|\n\xa7\xa1\x8f\xa9C\xa3mu*_A\x92\xf0\x13a\x00\xaa\x9c\xb4\x1b\xbe\x14\xbe\xb1\xf9\xfci/\xdf\xa0\x94'
             b'F\xdeZ\x9d\xde,\xa5\xf6\x8c\x1c\x0c!B\x92\x87\xcb-\xaa\xa3\xd2cu/s\xe0\x9f\xafDy\xd2\x81t)\xf6\x98\x00'
             b'\xaf\xdekY-\xc1\x98\x82\xbd\xf5\x81\xcc\xab\xf2\xcb\x91\x02\x9e\xf3\\L\xfd\xbb\xffI\xc1\xfa\x1b/\xe3\x1d'
             b'\xe7\xa5`\xec\xb4~\xbc\xfe2B[\x95o\x81\xb6\x99\x17H~;x\x91Q\xdb.x\xb1\xfd.\xbe~bk>\xa1e\xb4\xfb\x00\xcc'
             b'\xb7Q\xafPs)\xc4\xa3\x93\x9e\xa6\xdd\x9cP\xa0\xe78k\x01EykA\xafa\xf7\x85U\x94O;\xc2-\xc3\xbd\r\x00\xf8y'
             b'\x8aB\xb1\xaa\xa0\x83 e\x9a\xc79Z\xb4\xf3)')
    else:
        raise ValueError("The provided CA certificate is not valid!")
    root_key_exponent = 0x00010001
    cert_hash = SHA1.new(ca_cert.dump()[576:])
    public_key = RSA.construct((int.from_bytes(root_key_modulus), root_key_exponent))
    try:
        pkcs1_15.new(public_key).verify(cert_hash, ca_cert.signature)
        return True
    except ValueError:
        return False


def verify_cert_sig(ca_cert: Certificate, target_cert: Certificate) -> bool:
    """
    Verify a TMD or Ticket certificate using a CA certificate.

    Parameters
    ----------
    ca_cert: Certificate
        The CA certificate to use for verification.
    target_cert: Certificate
        The target certificate to verify.

    Returns
    -------
    bool
        Whether the certificate's signature is valid or not.
    """
    if ca_cert.issuer != "Root" or ca_cert.child_name.find("CA") == -1:
        raise ValueError("The provided certificate is not a CA certificate!")
    # The issuer of the TMD/Ticket certs is Root-CA0000000X, so prepend "Root-" to the CA cert child name. If these
    # don't match, then there's probably a mismatch between retail and development certs.
    if f"Root-{ca_cert.child_name}" != target_cert.issuer:
        raise ValueError("The certificate you are trying to verify does not match the provided CA certificate!")
    cert_hash = SHA1.new(target_cert.dump()[320:])
    public_key = RSA.construct((ca_cert.pub_key_modulus, ca_cert.pub_key_exponent))
    try:
        pkcs1_15.new(public_key).verify(cert_hash, target_cert.signature)
        return True
    except ValueError:
        return False


def verify_tmd_sig(tmd_cert: Certificate, tmd: TMD) -> bool:
    """
    Verify the signature of a TMD file using a TMD certificate.

    Parameters
    ----------
    tmd_cert: Certificate
        The TMD certificate to use for verification.
    tmd: TMD
        The TMD to verify.

    Returns
    -------
    bool
        Whether the TMD's signature is valid or not.
    """
    if tmd_cert.issuer.find("Root-CA") == -1 or tmd_cert.child_name.find("CP") == -1:
        raise ValueError("The provided TMD certificate is not valid!")
    if f"{tmd_cert.issuer}-{tmd_cert.child_name}" != tmd.signature_issuer:
        raise ValueError("The signature you are trying to verify was not created with the provided TMD certificate!")
    tmd_hash = SHA1.new(tmd.dump()[320:])
    public_key = RSA.construct((tmd_cert.pub_key_modulus, tmd_cert.pub_key_exponent))
    try:
        pkcs1_15.new(public_key).verify(tmd_hash, tmd.signature)
        return True
    except ValueError:
        return False


def verify_ticket_sig(ticket_cert: Certificate, ticket: Ticket) -> bool:
    """
    Verify the signature of a Ticket file using a Ticket certificate.

    Parameters
    ----------
    ticket_cert: Certificate
        The Ticket certificate to use for verification.
    ticket: Ticket
        The Ticket to verify.

    Returns
    -------
    bool
        Whether the Ticket's signature is valid or not.
    """
    if ticket_cert.issuer.find("Root-CA") == -1 or ticket_cert.child_name.find("XS") == -1:
        raise ValueError("The provided Ticket certificate is not valid!")
    if f"{ticket_cert.issuer}-{ticket_cert.child_name}" != ticket.signature_issuer:
        raise ValueError("The signature you are trying to verify was not created with the provided Ticket certificate!")
    ticket_hash = SHA1.new(ticket.dump()[320:])
    public_key = RSA.construct((ticket_cert.pub_key_modulus, ticket_cert.pub_key_exponent))
    try:
        pkcs1_15.new(public_key).verify(ticket_hash, ticket.signature)
        return True
    except ValueError:
        return False



# "title/nus.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki/NUS for details about the NUS

import requests
#import hashlib
from typing import List
#from urllib.parse import urlparse as _urlparse
from .title import Title
from .tmd import TMD
from .ticket import Ticket

_nus_endpoint = ["http://nus.cdn.shop.wii.com/ccs/download/", "http://ccs.cdn.wup.shop.nintendo.net/ccs/download/"]


def download_title(title_id: str, title_version: int = None, wiiu_endpoint: bool = False,
                   endpoint_override: str = None) -> Title:
    """
    Download an entire title and all of its contents, then load the downloaded components into a Title object for
    further use. This method is NOT recommended for general use, as it has absolutely no verbosity. It is instead
    recommended to call the individual download methods instead to provide more flexibility and output.

    Parameters
    ----------
    title_id : str
        The Title ID of the title to download.
    title_version : int, optional
        The version of the title to download. Defaults to latest if not set.
    wiiu_endpoint : bool, optional
        Whether the Wii U endpoint for the NUS should be used or not. This increases download speeds. Defaults to False.
    endpoint_override: str, optional
        A custom endpoint URL to use instead of the standard Wii or Wii U endpoints. Defaults to no override, and if
        set entirely overrides the "wiiu_endpoint" parameter.

    Returns
    -------
    Title
        A Title object containing all the data from the downloaded title.
    """
    # First, create the new title.
    title = Title()
    # Download and load the certificate chain, TMD, and Ticket.
    title.load_cert_chain(download_cert_chain(wiiu_endpoint, endpoint_override))
    title.load_tmd(download_tmd(title_id, title_version, wiiu_endpoint, endpoint_override))
    title.load_ticket(download_ticket(title_id, wiiu_endpoint, endpoint_override))
    # Download all contents
    title.load_content_records()
    title.content.content_list = download_contents(title_id, title.tmd, wiiu_endpoint, endpoint_override)
    # Return the completed title.
    return title


def download_tmd(title_id: str, title_version: int = None, wiiu_endpoint: bool = False,
                 endpoint_override: str = None) -> bytes:
    """
    Downloads the TMD of the Title specified in the object. Will download the latest version by default, or another
    version if it was manually specified in the object.

    Parameters
    ----------
    title_id : str
        The Title ID of the title to download the TMD for.
    title_version : int, option
        The version of the TMD to download. Defaults to latest if not set.
    wiiu_endpoint : bool, option
        Whether the Wii U endpoint for the NUS should be used or not. This increases download speeds. Defaults to False.
    endpoint_override: str, optional
        A custom endpoint URL to use instead of the standard Wii or Wii U endpoints. Defaults to no override, and if
        set entirely overrides the "wiiu_endpoint" parameter.

    Returns
    -------
    bytes
        The TMD file from the NUS.
    """
    # Build the download URL. The structure is download/<TID>/tmd for latest and download/<TID>/tmd.<version> for
    # when a specific version is requested.
    if endpoint_override is not None:
        endpoint_url = _validate_endpoint(endpoint_override)
    else:
        if wiiu_endpoint:
            endpoint_url = _nus_endpoint[1]
        else:
            endpoint_url = _nus_endpoint[0]
    tmd_url = endpoint_url + title_id + "/tmd"
    # Add the version to the URL if one was specified.
    if title_version is not None:
        tmd_url += "." + str(title_version)
    # Make the request.
    try:
        tmd_request = requests.get(url=tmd_url, headers={'User-Agent': 'wii libnup/1.0'}, stream=True)
    except requests.exceptions.ConnectionError:
        if endpoint_override:
            raise ValueError("A connection could not be made to the NUS endpoint. Please make sure that your endpoint "
                             "override is valid.")
        else:
            raise Exception("A connection could not be made to the NUS endpoint. The NUS may be unavailable.")
    # Handle a 404 if the TID/version doesn't exist.
    if tmd_request.status_code != 200:
        raise ValueError("The requested Title ID or TMD version does not exist. Please check the Title ID and Title"
                         " version and then try again.")
    # Save the raw TMD.
    raw_tmd = tmd_request.content
    # Use a TMD object to load the data and then return only the actual TMD.
    tmd_temp = TMD()
    tmd_temp.load(raw_tmd)
    tmd = tmd_temp.dump()
    return tmd


def download_ticket(title_id: str, wiiu_endpoint: bool = False, endpoint_override: str = None) -> bytes:
    """
    Downloads the Ticket of the Title specified in the object. This will only work if the Title ID specified is for
    a free title.

    Parameters
    ----------
    title_id : str
        The Title ID of the title to download the Ticket for.
    wiiu_endpoint : bool, option
        Whether the Wii U endpoint for the NUS should be used or not. This increases download speeds. Defaults to False.
    endpoint_override: str, optional
        A custom endpoint URL to use instead of the standard Wii or Wii U endpoints. Defaults to no override, and if
        set entirely overrides the "wiiu_endpoint" parameter.

    Returns
    -------
    bytes
        The Ticket file from the NUS.
    """
    # Build the download URL. The structure is download/<TID>/cetk, and cetk will only exist if this is a free
    # title.
    if endpoint_override is not None:
        endpoint_url = _validate_endpoint(endpoint_override)
    else:
        if wiiu_endpoint:
            endpoint_url = _nus_endpoint[1]
        else:
            endpoint_url = _nus_endpoint[0]
    ticket_url = endpoint_url + title_id + "/cetk"
    # Make the request.
    try:
        ticket_request = requests.get(url=ticket_url, headers={'User-Agent': 'wii libnup/1.0'}, stream=True)
    except requests.exceptions.ConnectionError:
        if endpoint_override:
            raise ValueError("A connection could not be made to the NUS endpoint. Please make sure that your endpoint "
                             "override is valid.")
        else:
            raise Exception("A connection could not be made to the NUS endpoint. The NUS may be unavailable.")
    if ticket_request.status_code != 200:
        raise ValueError("The requested Title ID does not exist, or refers to a non-free title. Tickets can only"
                         " be downloaded for titles that are free on the NUS.")
    # Save the raw cetk file.
    cetk = ticket_request.content
    # Use a Ticket object to load only the Ticket data from cetk and return it.
    ticket_temp = Ticket()
    ticket_temp.load(cetk)
    ticket = ticket_temp.dump()
    return ticket


def download_cert_chain(wiiu_endpoint: bool = False, endpoint_override: str = None) -> bytes:
    """
    Downloads the signing certificate chain used by all WADs. This uses System Menu 4.3U as the source.

    Parameters
    ----------
    wiiu_endpoint : bool, option
        Whether the Wii U endpoint for the NUS should be used or not. This increases download speeds. Defaults to False.
    endpoint_override: str, optional
        A custom endpoint URL to use instead of the standard Wii or Wii U endpoints. Defaults to no override, and if
        set entirely overrides the "wiiu_endpoint" parameter.

    Returns
    -------
    bytes
        The cert file.
    """
    # Download the TMD and cetk for System Menu 4.3U (v513).
    if endpoint_override is not None:
        endpoint_url = _validate_endpoint(endpoint_override)
    else:
        if wiiu_endpoint:
            endpoint_url = _nus_endpoint[1]
        else:
            endpoint_url = _nus_endpoint[0]
    tmd_url = endpoint_url + "0000000100000002/tmd.513"
    cetk_url = endpoint_url + "0000000100000002/cetk"
    try:
        tmd = requests.get(url=tmd_url, headers={'User-Agent': 'wii libnup/1.0'}, stream=True).content
        cetk = requests.get(url=cetk_url, headers={'User-Agent': 'wii libnup/1.0'}, stream=True).content
    except requests.exceptions.ConnectionError:
        if endpoint_override:
            raise ValueError("A connection could not be made to the NUS endpoint. Please make sure that your endpoint "
                             "override is valid.")
        else:
            raise Exception("A connection could not be made to the NUS endpoint. The NUS may be unavailable.")
    # Assemble the certificate chain.
    cert_chain = b''
    # Certificate Authority data.
    cert_chain += cetk[0x2A4 + 768:]
    # Certificate Policy (TMD certificate) data.
    cert_chain += tmd[0x328:0x328 + 768]
    # XS (Ticket certificate) data.
    cert_chain += cetk[0x2A4:0x2A4 + 768]
    # Since the cert chain is always the same, check the hash to make sure nothing went wildly wrong.
    # This is currently disabled because of the possibility that one may be downloading non-retail certs (gasp!).
    #if hashlib.sha1(cert_chain).hexdigest() != "ace0f15d2a851c383fe4657afc3840d6ffe30ad0":
    #    raise Exception("An unknown error has occurred downloading and creating the certificate.")
    return cert_chain


def download_content(title_id: str, content_id: int, wiiu_endpoint: bool = False,
                     endpoint_override: str = None) -> bytes:
    """
    Downloads a specified content for the title specified in the object.

    Parameters
    ----------
    title_id : str
        The Title ID of the title to download content from.
    content_id : int
        The Content ID of the content you wish to download.
    wiiu_endpoint : bool, option
        Whether the Wii U endpoint for the NUS should be used or not. This increases download speeds. Defaults to False.
    endpoint_override: str, optional
        A custom endpoint URL to use instead of the standard Wii or Wii U endpoints. Defaults to no override, and if
        set entirely overrides the "wiiu_endpoint" parameter.

    Returns
    -------
    bytes
        The downloaded content.
    """
    # Build the download URL. The structure is download/<TID>/<Content ID>.
    content_id_hex = hex(content_id)[2:]
    if len(content_id_hex) < 2:
        content_id_hex = "0" + content_id_hex
    if endpoint_override is not None:
        endpoint_url = _validate_endpoint(endpoint_override)
    else:
        if wiiu_endpoint:
            endpoint_url = _nus_endpoint[1]
        else:
            endpoint_url = _nus_endpoint[0]
    content_url = endpoint_url + title_id + "/000000" + content_id_hex
    # Make the request.
    try:
        content_request = requests.get(url=content_url, headers={'User-Agent': 'wii libnup/1.0'}, stream=True)
    except requests.exceptions.ConnectionError:
        if endpoint_override:
            raise ValueError("A connection could not be made to the NUS endpoint. Please make sure that your endpoint "
                             "override is valid.")
        else:
            raise Exception("A connection could not be made to the NUS endpoint. The NUS may be unavailable.")
    if content_request.status_code != 200:
        raise ValueError("The requested Title ID does not exist, or an invalid Content ID is present in the"
                         " content records provided.\n Failed while downloading Content ID: 000000" +
                         content_id_hex)
    content_data = content_request.content
    return content_data


def download_contents(title_id: str, tmd: TMD, wiiu_endpoint: bool = False,
                      endpoint_override: str = None) -> List[bytes]:
    """
    Downloads all the contents for the title specified in the object. This requires a TMD to already be available
    so that the content records can be accessed.

    Parameters
    ----------
    title_id : str
        The Title ID of the title to download content from.
    tmd : TMD
        The TMD that matches the title that the contents being downloaded are from.
    wiiu_endpoint : bool, option
        Whether the Wii U endpoint for the NUS should be used or not. This increases download speeds. Defaults to False.
    endpoint_override: str, optional
        A custom endpoint URL to use instead of the standard Wii or Wii U endpoints. Defaults to no override, and if
        set entirely overrides the "wiiu_endpoint" parameter.

    Returns
    -------
    List[bytes]
        A list of all the downloaded contents.
    """
    # Retrieve the content records from the TMD.
    content_records = tmd.content_records
    # Create a list of Content IDs to download.
    content_ids = []
    for content_record in content_records:
        content_ids.append(content_record.content_id)
    # Iterate over that list and download each content in it, then add it to the array of contents.
    content_list = []
    for content_id in content_ids:
        # Call self.download_content() for each Content ID.
        content = download_content(title_id, content_id, wiiu_endpoint, endpoint_override)
        content_list.append(content)
    return content_list


def _validate_endpoint(endpoint: str) -> str:
    """
    Validate the provided NUS endpoint URL and append the required path if necessary.

    Parameters
    ----------
    endpoint: str
        The NUS endpoint URL to validate.

    Returns
    -------
    str
        The validated NUS endpoint with the proper path.
    """
    # Find the root of the URL and then assemble the correct URL based on that.
    # TODO: Rewrite in a way that makes more sense and un-stub
    #new_url = _urlparse(endpoint)
    #if new_url.netloc == "":
    #    endpoint_url = "http://" + new_url.path + "/ccs/download/"
    #else:
    #    endpoint_url = "http://" + new_url.netloc + "/ccs/download/"
    return endpoint



# "__init__.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# These are the essential submodules from libWiiPy that you'd probably want imported by default.

__all__ = ["archive", "media", "nand", "title"]

from . import archive
from . import media
from . import nand
from . import title



# "title/commonkeys.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy

import binascii

common_key = 'ebe42a225e8593e448d9c5457381aaf7'
korean_key = '63b82bb4f4614e2e13f2fefbba4c9b7e'
vwii_key = '30bfc76e7c19afbb23163330ced7c28d'

development_key = 'a1604a6a7123b529ae8bec32c816fcaa'


def get_common_key(common_key_index, dev=False) -> bytes:
    """
    Gets the specified Wii Common Key based on the index provided. If an invalid common key index is provided, this
    function falls back on always returning key 0 (the Common Key). If the kwarg "dev" is specified, then key 0 will
    point to the development common key rather than the retail one. Keys 1 and 2 are unaffected by this argument.

    Possible values for common_key_index: 0: Common Key, 1: Korean Key, 2: vWii Key

    Parameters
    ----------
    common_key_index : int
        The index of the common key to be returned.
    dev : bool
        If the dev keys should be used in place of the retail keys. Only affects key 0.

    Returns
    -------
    bytes
        The specified common key, in binary format.
    """
    match common_key_index:
        case 0:
            if dev:
                common_key_bin = binascii.unhexlify(development_key)
            else:
                common_key_bin = binascii.unhexlify(common_key)
        case 1:
            common_key_bin = binascii.unhexlify(korean_key)
        case 2:
            common_key_bin = binascii.unhexlify(vwii_key)
        case _:
            common_key_bin = binascii.unhexlify(common_key)
    return common_key_bin



# "archive/lz77.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki/LZ77 for details about the LZ77 compression format.

import io
from dataclasses import dataclass as _dataclass
from typing import List as _List


_LZ_MIN_DISTANCE = 0x01   # Minimum distance for each reference.
_LZ_MAX_DISTANCE = 0x1000   # Maximum distance for each reference.
_LZ_MIN_LENGTH = 0x03   # Minimum length for each reference.
_LZ_MAX_LENGTH = 0x12   # Maximum length for each reference.


@_dataclass
class _LZNode:
    dist: int = 0
    len: int = 0
    weight: int = 0


def _compress_compare_bytes(buffer: _List[int], offset1: int, offset2: int, abs_len_max: int) -> int:
    # Compare bytes up to the maximum length we can match. Start by comparing the first 3 bytes, since that's the
    # minimum match length and this allows for a more optimized early exit.
    num_matched = 0
    while num_matched < abs_len_max:
        if buffer[offset1 + num_matched] != buffer[offset2 + num_matched]:
            break
        num_matched += 1
    return num_matched


def _compress_search_matches_optimized(buffer: _List[int], pos: int) -> (int, int):
    bytes_left = len(buffer) - pos
    global _LZ_MAX_DISTANCE, _LZ_MIN_LENGTH, _LZ_MAX_LENGTH, _LZ_MIN_DISTANCE
    # Default to only looking back 4096 bytes, unless we've moved fewer than 4096 bytes, in which case we should
    # only look as far back as we've gone.
    max_dist = min(_LZ_MAX_DISTANCE, pos)
    # Default to only matching up to 18 bytes, unless fewer than 18 bytes remain, in which case we can only match
    # up to that many bytes.
    max_len = min(_LZ_MAX_LENGTH, bytes_left)
    # Log the longest match we found and its offset.
    biggest_match, biggest_match_pos = 0, 0
    # Search for matches.
    for i in range(_LZ_MIN_DISTANCE, max_dist + 1):
        num_matched = _compress_compare_bytes(buffer, pos - i, pos, max_len)
        if num_matched > biggest_match:
            biggest_match = num_matched
            biggest_match_pos = i
            if biggest_match == max_len:
                break
    return biggest_match, biggest_match_pos


def _compress_search_matches_greedy(buffer: _List[int], pos: int) -> (int, int):
    # Finds and returns the first valid match, rather that finding the best one.
    bytes_left = len(buffer) - pos
    global _LZ_MAX_DISTANCE, _LZ_MAX_LENGTH, _LZ_MIN_DISTANCE
    # Default to only looking back 4096 bytes, unless we've moved fewer than 4096 bytes, in which case we should
    # only look as far back as we've gone.
    max_dist = min(_LZ_MAX_DISTANCE, pos)
    # Default to only matching up to 18 bytes, unless fewer than 18 bytes remain, in which case we can only match
    # up to that many bytes.
    max_len = min(_LZ_MAX_LENGTH, bytes_left)
    match, match_pos = 0, 0
    for i in range(_LZ_MIN_DISTANCE, max_dist + 1):
        match = _compress_compare_bytes(buffer, pos - i, pos, max_len)
        match_pos = i
        if match >= _LZ_MIN_LENGTH or match == max_len:
            break
    return match, match_pos


def _compress_node_is_ref(node: _LZNode) -> bool:
    return node.len >= _LZ_MIN_LENGTH


def _compress_get_node_cost(length: int) -> int:
    if length >= _LZ_MIN_LENGTH:
        num_bytes = 2
    else:
        num_bytes = 1
    return 1 + (num_bytes * 8)


def _compress_lz77_optimized(data: bytes) -> bytes:
    # Optimized compressor based around a node graph that finds optimal string matches. Originally the default
    # implementation, but unfortunately it's very slow.
    nodes = [_LZNode() for _ in range(len(data))]
    # Iterate over the uncompressed data, starting from the end.
    pos = len(data)
    global _LZ_MAX_LENGTH, _LZ_MIN_LENGTH, _LZ_MIN_DISTANCE
    data_list = list(data)
    while pos:
        pos -= 1
        node = nodes[pos]
        # Limit the maximum search length when we're near the end of the file.
        max_search_len = min(_LZ_MAX_LENGTH, len(data_list) - pos)
        if max_search_len < _LZ_MIN_DISTANCE:
            max_search_len = 1
        # Initialize as 1 for each, since that's all we could use if we weren't compressing.
        length, dist = 1, 1
        if max_search_len >= _LZ_MIN_LENGTH:
            length, dist = _compress_search_matches_optimized(data_list, pos)
        # Treat as direct bytes if it's too short to copy.
        if length == 0 or length < _LZ_MIN_LENGTH:
            length = 1
        # If the node goes to the end of the file, the weight is the cost of the node.
        if (pos + length) == len(data_list):
            node.len = length
            node.dist = dist
            node.weight = _compress_get_node_cost(length)
        # Otherwise, search for possible matches and determine the one with the best cost.
        else:
            weight_best = 0xFFFFFFFF  # This was originally UINT_MAX, but that isn't a thing here so 32-bit it is!
            len_best = 1
            while length:
                weight_next = nodes[pos + length].weight
                weight = _compress_get_node_cost(length) + weight_next
                if weight < weight_best:
                    len_best = length
                    weight_best = weight
                length -= 1
                if length != 0 and length < _LZ_MIN_LENGTH:
                    length = 1
            node.len = len_best
            node.dist = dist
            node.weight = weight_best
    # Write the compressed data.
    with io.BytesIO() as buffer:
        # Write the header data.
        buffer.write(b'LZ77\x10')  # The LZ type on the Wii is *always* 0x10.
        buffer.write(len(data).to_bytes(3, 'little'))

        src_pos = 0
        while src_pos < len(data):
            head = 0
            head_pos = buffer.tell()
            buffer.write(b'\x00')  # Reserve a byte for the chunk head.

            i = 0
            while i < 8 and src_pos < len(data):
                current_node = nodes[src_pos]
                length = current_node.len
                dist = current_node.dist
                # This is a reference node.
                if _compress_node_is_ref(current_node):
                    encoded = (((length - _LZ_MIN_LENGTH) & 0xF) << 12) | ((dist - _LZ_MIN_DISTANCE) & 0xFFF)
                    buffer.write(encoded.to_bytes(2))
                    head = (head | (1 << (7 - i))) & 0xFF
                # This is a direct copy node.
                else:
                    buffer.write(data[src_pos:src_pos + 1])
                src_pos += length
                i += 1

            pos = buffer.tell()
            buffer.seek(head_pos)
            buffer.write(head.to_bytes(1))
            buffer.seek(pos)

        buffer.seek(0)
        out_data = buffer.read()
    return out_data


def _compress_lz77_greedy(data: bytes) -> bytes:
    # Greedy compressor that processes the file start to end and saves the first matches found. Faster than the
    # optimized implementation, but creates larger files.
    global _LZ_MAX_LENGTH, _LZ_MIN_LENGTH, _LZ_MIN_DISTANCE
    with io.BytesIO() as buffer:
        # Write the header data.
        buffer.write(b'LZ77\x10')  # The LZ type on the Wii is *always* 0x10.
        buffer.write(len(data).to_bytes(3, 'little'))

        src_pos = 0
        data_list = list(data)
        while src_pos < len(data):
            head = 0
            head_pos = buffer.tell()
            buffer.write(b'\x00')  # Reserve a byte for the chunk head.

            i = 0
            while i < 8 and src_pos < len(data):
                length, dist = _compress_search_matches_greedy(data_list, src_pos)
                # This is a reference node.
                if length >= _LZ_MIN_LENGTH:
                    encoded = (((length - _LZ_MIN_LENGTH) & 0xF) << 12) | ((dist - _LZ_MIN_DISTANCE) & 0xFFF)
                    buffer.write(encoded.to_bytes(2))
                    head = (head | (1 << (7 - i))) & 0xFF
                    src_pos += length
                # This is a direct copy node.
                else:
                    buffer.write(data[src_pos:src_pos + 1])
                    src_pos += 1
                i += 1

            pos = buffer.tell()
            buffer.seek(head_pos)
            buffer.write(head.to_bytes(1))
            buffer.seek(pos)

        buffer.seek(0)
        out_data = buffer.read()
    return out_data


def compress_lz77(data: bytes, compression_level: int = 1) -> bytes:
    """
    Compresses data using the Wii's LZ77 compression algorithm and returns the compressed result. Supports two
    different levels of compression, one based around a "greedy" LZ compression algorithm and the other based around
    an optimized LZ compression algorithm. The greedy compressor, level 1, will produce a larger compressed file but
    will run noticeably faster than the optimized compressor, which is level 2, especially for larger data.

    Parameters
    ----------
    data: bytes
        The data to compress.
    compression_level: int
        The compression level to use, either 1 and 2. Default value is 1.

    Returns
    -------
    bytes
        The LZ77-compressed data.
    """
    if compression_level == 1:
        out_data = _compress_lz77_greedy(data)
    elif compression_level == 2:
        out_data = _compress_lz77_optimized(data)
    else:
        raise ValueError(f"Invalid compression level \"{compression_level}\"!\"")
    return out_data


def decompress_lz77(lz77_data: bytes) -> bytes:
    """
    Decompresses LZ77-compressed data and returns the decompressed result. Supports data both with and without the
    magic number 'LZ77' (which may not be present if the data is embedded in something else).

    Parameters
    ----------
    lz77_data: bytes
        The LZ77-compressed data to decompress.

    Returns
    -------
    bytes
        The decompressed data.
    """
    with io.BytesIO(lz77_data) as data:
        magic = data.read(4)
        # Assume if we didn't get the magic number that this data starts without it.
        if magic != b'LZ77':
            data.seek(0)
        # Other compression types are used by Nintendo, but only type 0x10 was used on the Wii.
        compression_type = int.from_bytes(data.read(1))
        if compression_type != 0x10:
            raise ValueError("This data is using an unsupported compression type!")
        decompressed_size = int.from_bytes(data.read(3), byteorder='little')
        # Use an integer list for storing decompressed data, this is much faster than using (and appending to) a
        # bytes object.
        out_data = [0] * decompressed_size
        pos = 0
        while pos < decompressed_size:
            flag = int.from_bytes(data.read(1))
            # Read bits in the flag from most to least significant.
            for x in range(7, -1, -1):
                # Avoids a buffer overrun if the final flag isn't fully used.
                if pos >= decompressed_size:
                    break
                # Result of 1, this means we're copying bytes from earlier in the data.
                if flag & (1 << x):
                    reference = int.from_bytes(data.read(2))
                    length = 3 + ((reference >> 12) & 0xF)
                    offset = pos - (reference & 0xFFF) - 1
                    for _ in range(length):
                        out_data[pos] = out_data[offset]
                        pos += 1
                        offset += 1
                        # Avoids a buffer overrun if the copy length would extend past the end of the file.
                        if pos >= decompressed_size:
                            break
                # Result of 0, use the next byte directly.
                else:
                    out_data[pos] = int.from_bytes(data.read(1))
                    pos += 1
        out_bytes = bytes(out_data)
        return out_bytes



# "nand/sys.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki//sys/uid.sys for information about uid.sys.

import io
import binascii
from typing import List
from dataclasses import dataclass as _dataclass


@_dataclass
class _UidSysEntry:
    """
    A _UidSysEntry object used to store an entry in uid.sys. Private class used by the sys module.

    Attributes
    ----------
    title_id : str
        The Title ID of the title this entry corresponds with.
    uid : int
        The UID assigned to the title this entry corresponds with.
    """
    title_id: str
    uid: int


class UidSys:
    """
    A UidSys object to parse and edit the uid.sys file stored in /sys/ on the Wii's NAND. This file is used to track all
    the titles that have been launched on a console.

    Attributes
    ----------
    uid_entries : List[_UidSysEntry]
        The entries stored in the uid.sys file.
    """

    def __init__(self):
        self.uid_entries: List[_UidSysEntry] = []

    def load(self, uid_sys: bytes) -> None:
        """
        Loads the raw data of uid.sys and parses it into a list of entries.

        Parameters
        ----------
        uid_sys : bytes
            The data of a uid.sys file.
        """
        # Sanity check to ensure the length is divisible by 12 bytes. If it isn't, then it is malformed.
        if (len(uid_sys) % 12) != 0:
            raise ValueError("The provided uid.sys appears to be corrupted!")
        entry_count = len(uid_sys) // 12
        with io.BytesIO(uid_sys) as uid_data:
            for i in range(entry_count):
                title_id = binascii.hexlify(uid_data.read(8)).decode()
                uid_data.seek(uid_data.tell() + 2)
                uid = int.from_bytes(uid_data.read(2))
                self.uid_entries.append(_UidSysEntry(title_id, uid))

    def dump(self) -> bytes:
        """
        Dumps the UidSys object back into a uid.sys file.

        Returns
        -------
        bytes
            The raw data of the uid.sys file.
        """
        uid_data = b''
        for record in self.uid_entries:
            uid_data += binascii.unhexlify(record.title_id.encode())
            uid_data += b'\x00' * 2
            uid_data += int.to_bytes(record.uid, 2)
        return uid_data

    def add(self, title_id: str | bytes) -> int:
        """
        Adds a new Title ID to the uid.sys file and returns the UID assigned to that title.

        Parameters
        ----------
        title_id : str, bytes
            The Title ID to add.

        Returns
        -------
        int
            The UID assigned to the new Title ID.
        """
        if type(title_id) is bytes:
            # This catches the format b'0000000100000002'
            if len(title_id) == 16:
                title_id_converted = title_id.encode()
            # This catches the format b'\x00\x00\x00\x01\x00\x00\x00\x02'
            elif len(title_id) == 8:
                title_id_converted = binascii.hexlify(title_id).decode()
            # If it isn't one of those lengths, it cannot possibly be valid, so reject it.
            else:
                raise ValueError("Title ID is not valid!")
        # Allow for a string like "0000000100000002"
        elif type(title_id) is str:
            if len(title_id) != 16:
                raise ValueError("Title ID is not valid!")
            title_id_converted = title_id
        else:
            raise TypeError("Title ID type is not valid! It must be either type str or bytes.")
        # Generate the new UID by incrementing the current highest UID by 1.
        try:
            new_uid = self.uid_entries[-1].uid + 1
        except IndexError:
            new_uid = 4096
        self.uid_entries.append(_UidSysEntry(title_id_converted, new_uid))
        return new_uid

    def create(self) -> None:
        """
        Creates a new uid.sys file and initializes it with the standard first entry of 1-2 with UID 4096. This allows
        for setting up a uid.sys file without having to load an existing one.
        """
        if len(self.uid_entries) != 0:
            raise Exception("A uid.sys file appears to already exist!")
        self.add("0000000100000002")



# "title/iospatcher.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# Module for applying patches to IOS WADs via a Title().

import io
from .title import Title


class IOSPatcher:
    """
    An IOSPatcher object that allows for applying patches to IOS WADs loaded into Title objects.

    Attributes
    ----------
    title : Title
        The loaded Title object to be patched.
    es_module_index : int
        The content index that ES resides in and where ES patches are applied.
    dip_module_index : int
        The content index that DIP resides in and where DIP patches are applied. -1 if DIP patches are not applied.
    """
    def __init__(self):
        self.title: Title = Title()
        self.es_module_index: int = -1
        self.dip_module_index: int = -1

    def load(self, title: Title) -> None:
        """
        Loads a Title object containing an IOS WAD and locates the content containing the ES module that needs to be
        patched.

        Parameters
        ----------
        title : Title
            A Title object containing the IOS to be patched.
        """
        # Check to ensure that this Title contains IOS. IOS always has a TID high of 00000001, and any TID low after
        # 00000002.
        tid = title.tmd.title_id
        if tid[:8] != "00000001" or tid[8:] == "00000001" or tid[8:] == "00000002":
            raise ValueError("This Title does not contain an IOS! Cannot load Title for patching.")

        # Now that we know this is IOS, we need to go ahead and check all of its contents until we find the one that
        # contains the ES module, since that's what we're patching.
        es_content_index = -1
        for content in range(len(title.content.content_records)):
            target_content = title.get_content_by_index(title.content.content_records[content].index)
            es_offset = target_content.find(b'\x45\x53\x3A')  # This is looking for "ES:"
            if es_offset != -1:
                es_content_index = title.content.content_records[content].index
                break

        # If we get here with no content index, then ES wasn't found. That probably means that this isn't IOS.
        if es_content_index == -1:
            raise Exception("ES module could not be found! Please ensure that this is an intact copy of an IOS.")

        self.title = title
        self.es_module_index = es_content_index

    def dump(self) -> Title:
        """
        Returns the patched Title object.

        Returns
        -------
        Title
            The patched Title object.
        """
        return self.title

    def patch_all(self) -> int:
        """
        Applies all patches to patch in fakesigning, ES_Identify access, /dev/flash access, and the version downgrading
        patch.

        Returns
        -------
        int
            The number of patches successfully applied.
        """
        patch_count = 0
        patch_count += self.patch_fakesigning()
        patch_count += self.patch_es_identify()
        patch_count += self.patch_nand_access()
        patch_count += self.patch_version_downgrading()
        return patch_count

    def patch_fakesigning(self) -> int:
        """
        Patches the trucha/fakesigning bug back into the IOS' ES module to allow it to accept fakesigned TMDs and
        Tickets.

        Returns
        -------
        int
            The number of patches successfully applied.
        """
        if self.es_module_index == -1:
            raise Exception("No valid IOS is loaded! Patching cannot continue.")

        target_content = self.title.get_content_by_index(self.es_module_index)

        patch_count = 0
        patch_sequences = [b'\x20\x07\x23\xa2', b'\x20\x07\x4b\x0b']
        for sequence in patch_sequences:
            start_offset = target_content.find(sequence)
            if start_offset != -1:
                with io.BytesIO(target_content) as content_data:
                    content_data.seek(start_offset + 1)
                    content_data.write(b'\x00')
                    content_data.seek(0)
                    target_content = content_data.read()
                    patch_count += 1

        self.title.set_content(target_content, self.es_module_index)

        return patch_count

    def patch_es_identify(self) -> int:
        """
        Patches the ability to call ES_Identify back into the IOS' ES module to allow for changing the permissions of a
        title.

        Returns
        -------
        int
            The number of patches successfully applied.
        """
        if self.es_module_index == -1:
            raise Exception("No valid IOS is loaded! Patching cannot continue.")

        target_content = self.title.get_content_by_index(self.es_module_index)

        patch_count = 0
        patch_sequence = b'\x28\x03\xd1\x23'
        start_offset = target_content.find(patch_sequence)
        if start_offset != -1:
            with io.BytesIO(target_content) as content_data:
                content_data.seek(start_offset + 2)
                content_data.write(b'\x00\x00')
                content_data.seek(0)
                target_content = content_data.read()
                patch_count += 1

        self.title.set_content(target_content, self.es_module_index)

        return patch_count

    def patch_nand_access(self) -> int:
        """
        Patches the ability to directly access /dev/flash back into the IOS' ES module to allow for raw access to the
        Wii's filesystem.

        Returns
        -------
        int
            The number of patches successfully applied.
        """
        if self.es_module_index == -1:
            raise Exception("No valid IOS is loaded! Patching cannot continue.")

        target_content = self.title.get_content_by_index(self.es_module_index)

        patch_count = 0
        patch_sequence = b'\x42\x8b\xd0\x01\x25\x66'
        start_offset = target_content.find(patch_sequence)
        if start_offset != -1:
            with io.BytesIO(target_content) as content_data:
                content_data.seek(start_offset + 2)
                content_data.write(b'\xe0')
                content_data.seek(0)
                target_content = content_data.read()
                patch_count += 1

        self.title.set_content(target_content, self.es_module_index)

        return patch_count

    def patch_version_downgrading(self) -> int:
        """
        Patches the ability to downgrade installed titles into IOS' ES module.

        Returns
        -------
        int
            The number of patches successfully applied.
        """
        if self.es_module_index == -1:
            raise Exception("No valid IOS is loaded! Patching cannot continue.")

        target_content = self.title.get_content_by_index(self.es_module_index)

        patch_count = 0
        patch_sequence = b'\xd2\x01\x4e\x56'
        start_offset = target_content.find(patch_sequence)
        if start_offset != -1:
            with io.BytesIO(target_content) as content_data:
                content_data.seek(start_offset)
                content_data.write(b'\xe0')
                content_data.seek(0)
                target_content = content_data.read()
                patch_count += 1

        self.title.set_content(target_content, self.es_module_index)

        return patch_count

    def patch_drive_inquiry(self) -> int:
        """
        Patches out IOS' drive inquiry on startup, allowing IOS to load without a disc drive. Only required/useful if
        you do not have a disc drive connected to your console.

        This drive inquiry patch is EXPERIMENTAL, and may introduce unexpected side effects on some consoles.

        Returns
        -------
        int
            The number of patches successfully applied.
        """
        if self.es_module_index == -1:
            raise Exception("No valid IOS is loaded! Patching cannot continue.")

        # This patch is applied to the DIP module rather than to ES, so we need to search the contents for the right one
        # first.
        for content in range(len(self.title.content.content_records)):
            target_content = self.title.get_content_by_index(self.title.content.content_records[content].index)
            dip_offset = target_content.find(b'\x44\x49\x50\x3a')  # This is looking for "DIP:"
            if dip_offset != -1:
                self.dip_module_index = self.title.content.content_records[content].index
                break

        # If we get here with no content index, then DIP wasn't found. That probably means that this isn't IOS.
        if self.dip_module_index == -1:
            raise Exception("DIP module could not be found! Please ensure that this is an intact copy of an IOS.")

        target_content = self.title.get_content_by_index(self.dip_module_index)

        patch_count = 0
        patch_sequence = b'\x49\x4c\x23\x90\x68\x0a'  # 49 4c 23 90 68 0a
        start_offset = target_content.find(patch_sequence)
        if start_offset != -1:
            with io.BytesIO(target_content) as content_data:
                content_data.seek(start_offset)
                content_data.write(b'\x20\x00\xe5\x38')
                content_data.seek(0)
                target_content = content_data.read()
                patch_count += 1

        self.title.set_content(target_content, self.dip_module_index)

        return patch_count



# "types.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy

from dataclasses import dataclass


@dataclass
class _ContentRecord:
    """
    A content record object that contains the details of a content contained in a title. This information must match
    the content stored at the index in the record, or else the content will not decrypt properly, as the hash of the
    decrypted data will not match the hash in the content record.

    Attributes
    ----------
    content_id : int
        The unique ID of the content.
    index : int
        The index of this content in the content records.
    content_type : int
        The type of the content.
    content_size : int
        The size of the content when decrypted.
    content_hash
        The SHA-1 hash of the decrypted content.
    """
    content_id: int
    index: int
    content_type: int  # Type of content, possible values of: 0x0001: Normal, 0x4001: DLC, 0x8001: Shared.
    content_size: int
    content_hash: bytes



# "nand/emunand.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# Code for handling setting up and modifying a Wii EmuNAND.

import os
import pathlib
import shutil
from dataclasses import dataclass as _dataclass
from typing import List
from ..title.ticket import Ticket
from ..title.title import Title
from ..title.tmd import TMD
from ..title.content import SharedContentMap as _SharedContentMap
from .sys import UidSys as _UidSys


class EmuNAND:
    """
    An EmuNAND object that allows for creating and modifying Wii EmuNANDs. Requires the path to the root of the
    EmuNAND, and can optionally take in a callback function to send logs to.

    Parameters
    ----------
    emunand_root : str, pathlib.Path
        The path to the EmuNAND root directory.
    callback : function
        A callback function to send EmuNAND logs to.

    Attributes
    ----------
    emunand_root : pathlib.Path
        The path to the EmuNAND root directory.
    """
    def __init__(self, emunand_root: str | pathlib.Path, callback: callable = None):
        self.emunand_root = pathlib.Path(emunand_root)
        self.log = callback if callback is not None else None

        self.import_dir = self.emunand_root.joinpath("import")
        self.meta_dir = self.emunand_root.joinpath("meta")
        self.shared1_dir = self.emunand_root.joinpath("shared1")
        self.shared2_dir = self.emunand_root.joinpath("shared2")
        self.sys_dir = self.emunand_root.joinpath("sys")
        self.ticket_dir = self.emunand_root.joinpath("ticket")
        self.title_dir = self.emunand_root.joinpath("title")
        self.tmp_dir = self.emunand_root.joinpath("tmp")
        self.wfs_dir = self.emunand_root.joinpath("wfs")

        self.import_dir.mkdir(exist_ok=True)
        self.meta_dir.mkdir(exist_ok=True)
        self.shared1_dir.mkdir(exist_ok=True)
        self.shared2_dir.mkdir(exist_ok=True)
        self.sys_dir.mkdir(exist_ok=True)
        self.ticket_dir.mkdir(exist_ok=True)
        self.title_dir.mkdir(exist_ok=True)
        self.tmp_dir.mkdir(exist_ok=True)
        self.wfs_dir.mkdir(exist_ok=True)

    def install_title(self, title: Title, skip_hash=False) -> None:
        """
        Install the provided Title object to the EmuNAND. This mimics a real WAD installation done by ES.

        This will create some system files required if they do not exist, but note that this alone is not enough for
        a working EmuNAND, other than for Dolphin which can fill in the gaps.

        Parameters
        ----------
        title : libWiiPy.title.Title
            The loaded Title object to install.
        skip_hash : bool, optional
            Skip the hash check and install the title regardless of its hashes. Defaults to false.
        """
        # Save the upper and lower portions of the Title ID, because these are used as target install directories.
        tid_upper = title.tmd.title_id[:8]
        tid_lower = title.tmd.title_id[8:]

        # Tickets are installed as <tid_lower>.tik in /ticket/<tid_upper>/
        ticket_dir = self.ticket_dir.joinpath(tid_upper)
        ticket_dir.mkdir(exist_ok=True)
        ticket_dir.joinpath(f"{tid_lower}.tik").write_bytes(title.ticket.dump())

        # The TMD and normal contents are installed to /title/<tid_upper>/<tid_lower>/content/, with the tmd being named
        # title.tmd and the contents being named <cid>.app.
        title_dir = self.title_dir.joinpath(tid_upper)
        title_dir.mkdir(exist_ok=True)
        title_dir = title_dir.joinpath(tid_lower)
        title_dir.mkdir(exist_ok=True)
        content_dir = title_dir.joinpath("content")
        if content_dir.exists():
            shutil.rmtree(content_dir)  # Clear the content directory so old contents aren't left behind.
        content_dir.mkdir(exist_ok=True)
        content_dir.joinpath("title.tmd").write_bytes(title.tmd.dump())
        for content_file in range(0, title.tmd.num_contents):
            if title.tmd.content_records[content_file].content_type == 1:
                content_file_name = f"{title.tmd.content_records[content_file].content_id:08X}".lower()
                content_dir.joinpath(f"{content_file_name}.app").write_bytes(
                    title.get_content_by_index(content_file, skip_hash=skip_hash))
        title_dir.joinpath("data").mkdir(exist_ok=True)  # Empty directory used for save data for the title.

        # Shared contents need to be installed to /shared1/, with incremental names determined by /shared1/content.map.
        content_map_path = self.shared1_dir.joinpath("content.map")
        content_map = _SharedContentMap()
        existing_hashes = []
        if content_map_path.exists():
            content_map.load(content_map_path.read_bytes())
            for record in content_map.shared_records:
                existing_hashes.append(record.content_hash)
        for content_file in range(0, title.tmd.num_contents):
            if title.tmd.content_records[content_file].content_type == 32769:
                if title.tmd.content_records[content_file].content_hash not in existing_hashes:
                    content_file_name = content_map.add_content(title.tmd.content_records[content_file].content_hash)
                    self.shared1_dir.joinpath(f"{content_file_name}.app").write_bytes(
                        title.get_content_by_index(content_file, skip_hash=skip_hash))
        self.shared1_dir.joinpath("content.map").write_bytes(content_map.dump())

        # The "footer" or meta file is installed as title.met in /meta/<tid_upper>/<tid_lower>/. Only write this if meta
        # is not nothing.
        meta_data = title.wad.get_meta_data()
        if meta_data != b'':
            meta_dir = self.meta_dir.joinpath(tid_upper)
            meta_dir.mkdir(exist_ok=True)
            meta_dir = meta_dir.joinpath(tid_lower)
            meta_dir.mkdir(exist_ok=True)
            meta_dir.joinpath("title.met").write_bytes(title.wad.get_meta_data())

        # Ensure we have a uid.sys file created.
        uid_sys_path = self.sys_dir.joinpath("uid.sys")
        uid_sys = _UidSys()
        if not uid_sys_path.exists():
            uid_sys.create()

    def uninstall_title(self, tid: str) -> None:
        """
        Uninstall the Title with the specified Title ID from the EmuNAND. This will leave shared contents unmodified.

        Parameters
        ----------
        tid : str
            The Title ID of the Title to uninstall.
        """
        # Save the upper and lower portions of the Title ID, because these are used as target install directories.
        tid_upper = tid[:8]
        tid_lower = tid[8:]

        if not self.title_dir.joinpath(tid_upper).joinpath(tid_lower).exists():
            raise ValueError(f"Title with Title ID {tid} does not appear to be installed!")

        # Begin by removing the Ticket, which is installed to /ticket/<tid_upper>/<tid_lower>.tik
        if self.ticket_dir.joinpath(tid_upper).joinpath(tid_lower + ".tik").exists():
            os.remove(self.ticket_dir.joinpath(tid_upper).joinpath(tid_lower + ".tik"))

        # The TMD and contents are stored in /title/<tid_upper>/<tid_lower>/. Remove the TMD and all contents, but don't
        # delete the entire directory if anything exists in data.
        title_dir = self.title_dir.joinpath(tid_upper).joinpath(tid_lower)
        if not title_dir.joinpath("data").exists():
            shutil.rmtree(title_dir)
        elif title_dir.joinpath("data").exists() and not os.listdir(title_dir.joinpath("data")):
            shutil.rmtree(title_dir)
        else:
            # There are files in data, so we only want to delete the content directory.
            shutil.rmtree(title_dir.joinpath("content"))

        # On the off chance this title has a meta entry, delete that too.
        if self.meta_dir.joinpath(tid_upper).joinpath(tid_lower).joinpath("title.met").exists():
            shutil.rmtree(self.meta_dir.joinpath(tid_upper).joinpath(tid_lower))

    @_dataclass
    class InstalledTitles:
        """
        An InstalledTitles object that is used to track a title type and any titles that belong to that type that are
        installed to an EmuNAND.

        Attributes
        ----------
        type : str
            The type (Title ID high) of the installed titles.
        titles : List[str]
            The Title ID low of each installed title.
        """
        type: str
        titles: List[str]

    def get_installed_titles(self) -> List[InstalledTitles]:
        """
        Scans for installed titles and returns a list of InstalledTitles objects, which each contain a title type
        (Title ID high) and a list of Title ID lows that are installed under it.

        Returns
        -------
        List[InstalledTitles]
            The titles installed to the EmuNAND.
        """
        # Scan for TID highs present.
        tid_highs = [d for d in self.title_dir.iterdir() if d.is_dir()]
        # Iterate through each one, verify that every TID low directory contains a TMD, and then add it to the list.
        installed_titles = []
        for high in tid_highs:
            tid_lows = [d for d in high.iterdir() if d.is_dir()]
            valid_lows = []
            for low in tid_lows:
                if low.joinpath("content", "title.tmd").exists():
                    valid_lows.append(low.name.upper())
            installed_titles.append(self.InstalledTitles(high.name.upper(), valid_lows))
        return installed_titles

    def get_title_tmd(self, tid: str) -> TMD:
        """
        Gets the TMD for a title installed to the EmuNAND, and returns it as a TMD objects. Returns an error if the
        TMD for the specified Title ID does not exist.

        Parameters
        ----------
        tid : str
            The Title ID of the Title to get the TMD for.

        Returns
        -------
        TMD
            The TMD for the Title.
        """
        # Validate the TID, then build a path to the TMD file to verify that it exists.
        if len(tid) != 16:
            raise ValueError(f"Title ID \"{tid}\" is not a valid!")
        tid_high = tid[:8].lower()
        tid_low = tid[8:].lower()
        tmd_path = self.title_dir.joinpath(tid_high, tid_low, "content", "title.tmd")
        if not tmd_path.exists():
            raise FileNotFoundError(f"Title with Title ID {tid} does not appear to be installed!")
        tmd = TMD()
        tmd.load(tmd_path.read_bytes())
        return tmd

    def get_title_ticket(self, tid: str) -> Ticket:
        """
        Gets the Ticket for a title installed to the EmuNAND, and returns it as a Ticket object. Returns an error if
        the Ticket for the specified Title ID does not exist.

        Parameters
        ----------
        tid : str
            The Title ID of the Title to get the Ticket for.

        Returns
        -------
        Ticket
            The Ticket for the Title.
        """
        # Validate the TID, then build a path to the Ticket files to verify that it exists.
        if len(tid) != 16:
            raise ValueError(f"Title ID \"{tid}\" is not a valid!")
        tid_high = tid[:8].lower()
        tid_low = tid[8:].lower()
        ticket_path = self.ticket_dir.joinpath(tid_high, f"{tid_low}.tik")
        if not ticket_path.exists():
            raise FileNotFoundError(f"No Ticket exists for the title with Title ID {tid}!")
        ticket = Ticket()
        ticket.load(ticket_path.read_bytes())
        return ticket



# "title/util.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# General title-related utilities that don't fit within a specific module.

import math
from ..shared import _wii_menu_versions, _vwii_menu_versions


def title_ver_dec_to_standard(version: int, title_id: str, vwii: bool = False) -> str:
    """
    Converts a title's version from decimal form (vXXX, the way the version is stored in the TMD/Ticket) to its standard
    and human-readable form (vX.X). The Title ID is required as some titles handle this version differently from others.
    For the System Menu, the returned version will include the region code (ex. 4.3U).

    Parameters
    ----------
    version : int
        The version of the title, in decimal form.
    title_id : str
        The Title ID that the version is associated with.
    vwii : bool
        Whether this title is for the vWii or not. Only relevant for the System Menu.

    Returns
    -------
    str
        The version of the title, in standard form.
    """
    version_out = ""
    if title_id == "0000000100000002":
        if vwii:
            try:
                version_out = list(_vwii_menu_versions.keys())[list(_vwii_menu_versions.values()).index(version)]
            except ValueError:
                version_out = ""
        else:
            try:
                version_out = list(_wii_menu_versions.keys())[list(_wii_menu_versions.values()).index(version)]
            except ValueError:
                version_out = ""
    else:
        # For most channels, we need to get the floored value of version / 256 for the major version, and the version %
        # 256 as the minor version. Minor versions > 9 are intended, as Nintendo themselves frequently used them.
        version_upper = math.floor(version / 256)
        version_lower = version % 256
        version_out = f"{version_upper}.{version_lower}"

    return version_out


def title_ver_standard_to_dec(version: str, title_id: str) -> int:
    """
    Converts a title's version from its standard and human-readable form (vX.X) to its decimal form (vXXX, the way the
    version is stored in the TMD/Ticket). The Title ID is required as some titles handle this version differently from
    others. For the System Menu, the supplied version must include the region code (ex. 4.3U) for the conversion to
    work correctly.

    Parameters
    ----------
    version : str
        The version of the title, in standard form.
    title_id : str
        The Title ID that the version is associated with.

    Returns
    -------
    int
        The version of the title, in decimal form.
    """
    version_out = 0
    if title_id == "0000000100000002":
        raise ValueError("The System Menu's version cannot currently be converted.")
    else:
        version_str_split = version.split(".")
        version_upper = int(version_str_split[0]) * 256
        version_lower = int(version_str_split[1])
        version_out = version_upper + version_lower

    return version_out



# "title/content.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki/Title for details about how titles are formatted

import binascii
import io
import hashlib
from typing import List
from dataclasses import dataclass as _dataclass
from enum import IntEnum as _IntEnum
from ..types import _ContentRecord
from ..shared import _pad_bytes, _align_value
from .crypto import decrypt_content, encrypt_content


class ContentType(_IntEnum):
    NORMAL = 1
    HASH_TREE = 3
    DLC = 16385
    SHARED = 32769


class ContentRegion:
    """
    A ContentRegion object to parse the continuous content region of a WAD. Allows for retrieving content from the
    region in both encrypted or decrypted form, and setting new content.

    Attributes
    ----------
    content_records : List[_ContentRecord]
        The content records for the content stored in the region.
    num_contents : int
        The total number of contents stored in the region.
    """

    def __init__(self):
        self.content_records: List[_ContentRecord] = []
        self.content_region_size: int = 0  # Size of the content region.
        self.num_contents: int = 0  # Number of contents in the content region.
        self.content_start_offsets: List[int] = [0]  # The start offsets of each content in the content region.
        self.content_list: List[bytes] = []

    def load(self, content_region: bytes, content_records: List[_ContentRecord]) -> None:
        """
        Loads the raw content region and builds a list of all the contents.

        Parameters
        ----------
        content_region : bytes
            The raw data for the content region being loaded.
        content_records : list[_ContentRecord]
            A list of ContentRecord objects detailing all contents contained in the region.
        """
        self.content_records = content_records
        # Get the total size of the content region.
        self.content_region_size = len(content_region)
        with io.BytesIO(content_region) as content_region_data:
            self.num_contents = len(self.content_records)
            # Calculate the offsets of each content in the content region.
            # Content is aligned to 16 bytes, however a new content won't start until the next multiple of 64 bytes.
            # Because of this, we need to add bytes to the next 64 byte offset if the previous content wasn't that long.
            for content in self.content_records[:-1]:
                start_offset = content.content_size + self.content_start_offsets[-1]
                if (content.content_size % 64) != 0:
                    start_offset += 64 - (content.content_size % 64)
                self.content_start_offsets.append(start_offset)
            # Build a list of all the encrypted content data.
            for content in range(self.num_contents):
                # Seek to the start of the content based on the list of offsets.
                content_region_data.seek(self.content_start_offsets[content])
                # Calculate the number of bytes we need to read by adding bytes up the nearest multiple of 16 if needed.
                bytes_to_read = self.content_records[content].content_size
                if (bytes_to_read % 16) != 0:
                    bytes_to_read += 16 - (bytes_to_read % 16)
                # Read the file based on the size of the content in the associated record, then append that data to
                # the list of content.
                content_enc = content_region_data.read(bytes_to_read)
                self.content_list.append(content_enc)

    def dump(self) -> tuple[bytes, int]:
        """
        Takes the list of contents and assembles them back into one content region. Returns this content region as a
        bytes object and sets the raw content region variable to this result, then calls load() again to make sure the
        content list matches the raw data.

        Returns
        -------
        bytes
            The full ContentRegion as bytes, including padding between content.
        int
            The size of the ContentRegion, including padding.
        """
        content_region_data = b''
        for content in self.content_list:
            # If this isn't the first content, pad the whole region to 64 bytes before the next one.
            if content_region_data != b'':
                content_region_data = _pad_bytes(content_region_data, 64)
            # Calculate padding after this content before the next one.
            padding_bytes = 0
            if (len(content) % 16) != 0:
                padding_bytes = 16 - (len(content) % 16)
            # Write content data, then the padding afterward if necessary.
            content_region_data += content
            if padding_bytes > 0:
                content_region_data += b'\x00' * padding_bytes
        # Calculate the size of the whole content region.
        content_region_size = 0
        for record in range(len(self.content_records)):
            if record is len(self.content_records) - 1:
                content_region_size += self.content_records[record].content_size
            else:
                content_region_size += _align_value(self.content_records[record].content_size, 64)
        return content_region_data, content_region_size

    def get_enc_content_by_index(self, index: int) -> bytes:
        """
        Gets an individual content from the content region based on the provided index, in encrypted form.

        Parameters
        ----------
        index : int
            The index of the content you want to get.

        Returns
        -------
        bytes
            The encrypted content listed in the content record.
        """
        if index >= self.num_contents:
            raise ValueError(f"You are trying to get the content at index {index}, but no content with that "
                             f"index exists!")
        content_enc = self.content_list[index]
        return content_enc

    def get_enc_content_by_cid(self, cid: int) -> bytes:
        """
        Gets an individual content from the content region based on the provided Content ID, in encrypted form.

        Parameters
        ----------
        cid : int
            The Content ID of the content you want to get. Expected to be in decimal form, not hex.

        Returns
        -------
        bytes
            The encrypted content listed in the content record.
        """
        try:
            content_index = self.get_index_from_cid(cid)
        except ValueError:
            raise ValueError(f"You are trying to get a content with Content ID {cid}, "
                             f"but no content with that ID exists!")
        content_enc = self.get_enc_content_by_index(content_index)
        return content_enc

    def get_enc_contents(self) -> List[bytes]:
        """
        Gets a list of all encrypted contents from the content region.

        Returns
        -------
        List[bytes]
            A list containing all encrypted contents.
        """
        return self.content_list

    def get_content_by_index(self, index: int, title_key: bytes, skip_hash=False) -> bytes:
        """
        Gets an individual content from the content region based on the provided index, in decrypted form.

        Parameters
        ----------
        index : int
            The index of the content you want to get.
        title_key : bytes
            The Title Key for the title the content is from.
        skip_hash : bool, optional
            Skip the hash check and return the content regardless of its hash. Defaults to false.

        Returns
        -------
        bytes
            The decrypted content listed in the content record.
        """
        # Get the content index in the Content Record to ensure decryption works properly.
        cnt_index = self.content_records[index].index
        content_enc = self.get_enc_content_by_index(index)
        content_dec = decrypt_content(content_enc, title_key, cnt_index, self.content_records[index].content_size)
        # Hash the decrypted content and ensure that the hash matches the one in its Content Record.
        # If it does not, then something has gone wrong in the decryption, and an error will be thrown.
        content_dec_hash = hashlib.sha1(content_dec).hexdigest()
        content_record_hash = str(self.content_records[index].content_hash.decode())
        # Compare the hash and throw a ValueError if the hash doesn't match.
        if content_dec_hash != content_record_hash:
            if skip_hash:
                print("Ignoring hash mismatch for content index " + str(index))
            else:
                raise ValueError("Content hash did not match the expected hash in its record! The incorrect Title Key "
                                 "may have been used!\n"
                                 "Expected hash is: {}\n".format(content_record_hash) +
                                 "Actual hash is: {}".format(content_dec_hash))
        return content_dec

    def get_content_by_cid(self, cid: int, title_key: bytes, skip_hash=False) -> bytes:
        """
        Gets an individual content from the content region based on the provided Content ID, in decrypted form.

        Parameters
        ----------
        cid : int
            The Content ID of the content you want to get. Expected to be in decimal form, not hex.
        title_key : bytes
            The Title Key for the title the content is from.
        skip_hash : bool, optional
            Skip the hash check and return the content regardless of its hash. Defaults to false.

        Returns
        -------
        bytes
            The decrypted content listed in the content record.
        """
        try:
            content_index = self.get_index_from_cid(cid)
        except ValueError:
            raise ValueError(f"You are trying to get a content with Content ID {cid}, "
                             f"but no content with that ID exists!")
        content_dec = self.get_content_by_index(content_index, title_key, skip_hash)
        return content_dec

    def get_contents(self, title_key: bytes, skip_hash=False) -> List[bytes]:
        """
        Gets a list of all contents from the content region, in decrypted form.

        Parameters
        ----------
        title_key : bytes
            The Title Key for the title the content is from.
        skip_hash : bool, optional
            Skip the hash check and return the content regardless of its hash. Defaults to false.

        Returns
        -------
        List[bytes]
            A list containing all decrypted contents.
        """
        dec_contents: List[bytes] = []
        # Iterate over every content, get the decrypted version of it, then add it to a list and return it.
        for content in range(self.num_contents):
            dec_contents.append(self.get_content_by_index(content, title_key, skip_hash))
        return dec_contents

    def get_index_from_cid(self, cid: int) -> int:
        """
        Gets the index of a content by its Content ID.

        Parameters
        ----------
        cid : int
            The Content ID to get the index of.

        Returns
        -------
        int
            The content index.
        """
        # Get a list of the current Content IDs, so we can make sure the target one exists.
        content_ids = []
        for record in self.content_records:
            content_ids.append(record.content_id)
        if cid not in content_ids:
            raise ValueError("The specified Content ID does not exist!")
        index = content_ids.index(cid)
        return index

    def add_enc_content(self, enc_content: bytes, cid: int, index: int, content_type: int, content_size: int,
                        content_hash: bytes) -> None:
        """
        Adds a new encrypted content to the ContentRegion, and adds the provided Content ID, index, content type,
        content size, and content hash to a new record in the ContentRecord list.

        Parameters
        ----------
        enc_content : bytes
            The new encrypted content to add.
        cid : int
            The Content ID to assign the new content in the content record.
        index : int
            The index used when encrypting the new content.
        content_type : int
            The type of the new content.
        content_size : int
            The size of the new encrypted content when decrypted.
        content_hash : bytes
            The hash of the new encrypted content when decrypted.
        """
        # Check to make sure this isn't reusing an already existing Content ID or index first.
        for record in self.content_records:
            if record.content_id == cid:
                raise ValueError("Content with a Content ID of " + str(cid) + " already exists!")
            elif record.index == index:
                raise ValueError("Content with an index of " + str(index) + " already exists!")
        # If we're good, then append all the data and create a new ContentRecord().
        self.content_list.append(enc_content)
        self.content_records.append(_ContentRecord(cid, index, content_type, content_size, content_hash))
        self.num_contents += 1

    def add_content(self, dec_content: bytes, cid: int, content_type: int, title_key: bytes) -> None:
        """
        Adds a new decrypted content to the end of the ContentRegion, and adds the provided Content ID, content type,
        content size, and content hash to a new record in the ContentRecord list. The index will be automatically
        assigned by incrementing the current highest index in the records.

        This first gets the content hash and size from the provided data, and then encrypts the content with the
        provided Title Key before adding it to the ContentRegion.

        Parameters
        ----------
        dec_content : bytes
            The new decrypted content to add.
        cid : int
            The Content ID to assign the new content in the content record.
        content_type : int
            The type of the new content.
        title_key : bytes
            The Title Key that matches the other content in the ContentRegion.
        """
        # Find the current highest content index and increment it for this content.
        content_indices = []
        for record in self.content_records:
            content_indices.append(record.index)
        index = max(content_indices) + 1
        content_size = len(dec_content)
        content_hash = str.encode(hashlib.sha1(dec_content).hexdigest())
        enc_content = encrypt_content(dec_content, title_key, index)
        self.add_enc_content(enc_content, cid, index, content_type, content_size, content_hash)

    def set_enc_content(self, enc_content: bytes, index: int, content_size: int, content_hash: bytes, cid: int = None,
                        content_type: int = None) -> None:
        """
        Sets the content at the provided content index to the provided new encrypted content. The provided hash and
        content size are set in the corresponding content record. A new Content ID or content type can also be
        specified, but if it isn't then the current values are preserved.

        Parameters
        ----------
        enc_content : bytes
            The new encrypted content to set.
        index : int
            The target index to set the new content at.
        content_size : int
            The size of the new encrypted content when decrypted.
        content_hash : bytes
            The hash of the new encrypted content when decrypted.
        cid : int, optional
            The Content ID to assign the new content in the content record. Current value will be preserved if not set.
        content_type : int, optional
            The type of the new content. Current value will be preserved if not set.
        """
        if index >= self.num_contents:
            raise ValueError(f"You are trying to set the content at index {index}, but no content with that "
                             f"index currently exists!")
        # Reassign the values, but only set the optional ones if they were passed.
        self.content_records[index].content_size = content_size
        self.content_records[index].content_hash = content_hash
        if cid is not None:
            self.content_records[index].content_id = cid
        if content_type is not None:
            self.content_records[index].content_type = content_type
        # Add blank entries to the list to ensure that its length matches the length of the content record list.
        while len(self.content_list) < len(self.content_records):
            self.content_list.append(b'')
        self.content_list[index] = enc_content

    def set_content(self, dec_content: bytes, index: int, title_key: bytes, cid: int = None,
                    content_type: int = None) -> None:
        """
        Sets the content at the provided content index to the provided new decrypted content. The hash and content size
        of this content will be generated and then set in the corresponding content record. A new Content ID or content
        type can also be specified, but if it isn't then the current values are preserved.

        The provided Title Key is used to encrypt the content so that it can be set in the ContentRegion.

        Parameters
        ----------
        dec_content : bytes
            The new decrypted content to set.
        index : int
            The index to place the new content at.
        title_key : bytes
            The Title Key that matches the new decrypted content.
        cid : int
            The Content ID to assign the new content in the content record.
        content_type : int
            The type of the new content.
        """
        # Store the size of the new content.
        content_size = len(dec_content)
        # Calculate the hash of the new content.
        content_hash = str.encode(hashlib.sha1(dec_content).hexdigest())
        # Encrypt the content using the provided Title Key and the index from the Content Record, to ensure that
        # encryption will succeed even if the provided index doesn't match the content's index.
        enc_content = encrypt_content(dec_content, title_key, self.content_records[index].index)
        # Pass values to set_enc_content()
        self.set_enc_content(enc_content, index, content_size, content_hash, cid, content_type)

    def load_enc_content(self, enc_content: bytes, index: int) -> None:
        """
        Loads the provided encrypted content into the content region at the specified index, with the assumption that
        it matches the record at that index. Not recommended for most use cases, use decrypted content and
        load_content() instead.

        Parameters
        ----------
        enc_content : bytes
            The encrypted content to load.
        index : int
            The content index to load the content at.
        """
        if index >= self.num_contents:
            raise ValueError(f"You are trying to load the content at index {index}, but no content with that "
                             f"index currently exists! Make sure the correct content records have been loaded.")
        # Add blank entries to the list to ensure that its length matches the length of the content record list.
        while len(self.content_list) < len(self.content_records):
            self.content_list.append(b'')
        self.content_list[index] = enc_content

    def load_content(self, dec_content: bytes, index: int, title_key: bytes) -> None:
        """
        Loads the provided decrypted content into the ContentRegion at the specified index, but first checks to make
        sure that it matches the corresponding record. This content will then be encrypted using the provided Title Key
        before being loaded.

        Parameters
        ----------
        dec_content : bytes
            The decrypted content to load.
        index : int
            The index to load the content at.
        title_key: bytes
            The Title Key that matches the decrypted content.
        """
        if index >= self.num_contents:
            raise ValueError(f"You are trying to load the content at index {index}, but no content with that "
                             f"index currently exists! Make sure the correct content records have been loaded.")
        # Check the hash of the content against the hash stored in the record to ensure it matches.
        content_hash = hashlib.sha1(dec_content).hexdigest()
        if content_hash != self.content_records[index].content_hash.decode():
            raise ValueError("The decrypted content provided does not match the record at the provided index. \n"
                             "Expected hash is: {}\n".format(self.content_records[index].content_hash.decode()) +
                             "Actual hash is: {}".format(content_hash))
        # Add blank entries to the list to ensure that its length matches the length of the content record list.
        while len(self.content_list) < len(self.content_records):
            self.content_list.append(b'')
        # If the hash matches, encrypt the content and set it where it belongs.
        # This uses the index from the content records instead of just the index given, because there are some poorly
        # made custom WADs out there that don't have the contents in order, for whatever reason.
        enc_content = encrypt_content(dec_content, title_key, self.content_records[index].index)
        self.content_list[index] = enc_content

    def remove_content_by_index(self, index: int) -> None:
        """
        Removes the content at the specified index from the ContentRegion and content records.

        This will allow gaps to be left in content indices, however this should not cause any issues.

        Parameters
        ----------
        index : int
            The index of the content you want to remove.
        """
        if index >= self.num_contents:
            raise ValueError(f"You are trying to remove the content at index {index}, but no content with "
                             f"that index currently exists!")
        # Delete the target index from both the content list and content records.
        self.content_list.pop(index)
        self.content_records.pop(index)
        self.num_contents -= 1

    def remove_content_by_cid(self, cid: int) -> None:
        """
        Removes the content with the specified Content ID from the ContentRegion and content records.

        This will allow gaps to be left in content indices, however this should not cause any issues.

        Parameters
        ----------
        cid : int
            The Content ID of the content you want to remove.
        """
        try:
            index = self.get_index_from_cid(cid)
        except ValueError:
            raise ValueError(f"You are trying to remove content with Content ID {cid}, "
                             f"but no content with that ID exists!")
        self.remove_content_by_index(index)


@_dataclass
class _SharedContentRecord:
    """
    A _SharedContentRecord object used to store the data of a specific content stored in /shared1/. Private class used
    by the content module.

    Attributes
    ----------
    shared_id : str
        The incremental ID used to store the shared content.
    content_hash : bytes
        The SHA-1 hash of the shared content.
    """
    shared_id: str
    content_hash: bytes


class SharedContentMap:
    """
    A SharedContentMap object to parse and edit the content.map file stored in /shared1/ on the Wii's NAND. This file is
    used to keep track of all shared contents installed on the console.

    Attributes
    ----------
    shared_records : List[_SharedContentRecord]
        The shared content records stored in content.map.
    """

    def __init__(self):
        self.shared_records: List[_SharedContentRecord] = []

    def load(self, content_map: bytes) -> None:
        """
        Loads the raw content map and parses the records in it.

        Parameters
        ----------
        content_map : bytes
            The data of a content.map file.
        """
        # Sanity check to ensure the length is divisible by 28 bytes. If it isn't, then it is malformed.
        if (len(content_map) % 28) != 0:
            raise ValueError("The provided content map appears to be corrupted!")
        entry_count = len(content_map) // 28
        with io.BytesIO(content_map) as map_data:
            for i in range(entry_count):
                shared_id = str(map_data.read(8).decode())
                content_hash = binascii.hexlify(map_data.read(20))
                self.shared_records.append(_SharedContentRecord(shared_id, content_hash))

    def dump(self) -> bytes:
        """
        Dumps the SharedContentMap object back into a content.map file.

        Returns
        -------
        bytes
            The raw data of the content.map file.
        """
        map_data = b''
        for record in self.shared_records:
            map_data += record.shared_id.encode()
            map_data += binascii.unhexlify(record.content_hash)
        return map_data

    def add_content(self, content_hash: str | bytes) -> str:
        """
        Adds a new shared content SHA-1 hash to the content map and returns the file name assigned to that hash.

        Parameters
        ----------
        content_hash : str, bytes
            The SHA-1 hash of the new shared content.

        Returns
        -------
        str
            The filename assigned to the provided content hash.
        """
        if type(content_hash) is bytes:
            # This catches the format b'GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG'
            if len(content_hash) == 40:
                content_hash_converted = content_hash
            # This catches the format
            # b'\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG\xGG'
            elif len(content_hash) == 20:
                content_hash_converted = binascii.hexlify(content_hash)
            # If it isn't one of those lengths, it cannot possibly be valid, so reject it.
            else:
                raise ValueError("SHA-1 hash is not valid!")
        # Allow for a string like "GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG"
        elif type(content_hash) is str:
            content_hash_converted = content_hash.encode()
        # If the hash isn't bytes or a string, it isn't valid and is rejected.
        else:
            raise TypeError("SHA-1 hash type is not valid! It must be either type str or bytes.")

        # Generate the file name for the new shared content by incrementing the highest name by 1. Thank you, Nintendo,
        # for not just storing these as integers like you did EVERYWHERE else.
        try:
            maximum_index = int(self.shared_records[-1].shared_id, 16)
            new_index = f"{maximum_index + 1:08X}".lower()
        except IndexError:
            new_index = f"{0:08X}"
        self.shared_records.append(_SharedContentRecord(new_index, content_hash_converted))
        return new_index



# "shared.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# This file defines general functions that may be useful in other modules of libWiiPy. Putting them here cuts down on
# clutter in other files.


def _align_value(value, alignment=64) -> int:
    """
    Aligns the provided value to the set alignment (defaults to 64). Private function used by other libWiiPy modules.

    Parameters
    ----------
    value : int
        The value to align.
    alignment : int
        The number to align to. Defaults to 64.

    Returns
    -------
    int
        The aligned value.
    """
    if (value % alignment) != 0:
        aligned_value = value + (alignment - (value % alignment))
        return aligned_value
    return value


def _pad_bytes(data, alignment=64) -> bytes:
    """
    Pads the provided bytes object to the provided alignment (defaults to 64). Private function used by other libWiiPy
    modules.

    Parameters
    ----------
    data : bytes
        The data to align.
    alignment : int
        The number to align to. Defaults to 64.

    Returns
    -------
    bytes
        The aligned data.
    """
    while (len(data) % alignment) != 0:
        data += b'\x00'
    return data


def _bitmask(x: int) -> int:
    return 1 << x


_wii_menu_versions = {
    "Prelaunch": [0, 1, 2],
    "1.0J": 64,
    "1.0U": 33,
    "1.0E": 34,
    "2.0J": 128,
    "2.0U": 97,
    "2.0E": 130,
    "2.1E": 162,
    "2.2J": 192,
    "2.2U": 193,
    "2.2E": 194,
    "3.0J": 224,
    "3.0U": 225,
    "3.0E": 226,
    "3.1J": 256,
    "3.1U": 257,
    "3.1E": 258,
    "3.2J": 288,
    "3.2U": 289,
    "3.2E": 290,
    "3.3J": 352,
    "3.3U": 353,
    "3.3E": 354,
    "3.3K": 326,
    "3.4J": 384,
    "3.4U": 385,
    "3.4E": 386,
    "3.5K": 390,
    "4.0J": 416,
    "4.0U": 417,
    "4.0E": 418,
    "4.1J": 448,
    "4.1U": 449,
    "4.1E": 450,
    "4.1K": 454,
    "4.2J": 480,
    "4.2U": 481,
    "4.2E": 482,
    "4.2K": 486,
    "4.3J": 512,
    "4.3U": 513,
    "4.3E": 514,
    "4.3K": 518,
    "4.3U-Mini": 4609,
    "4.3E-Mini": 4610
}


_vwii_menu_versions = {
    "vWii-1.0.0J": 512,
    "vWii-1.0.0U": 513,
    "vWii-1.0.0E": 514,
    "vWii-4.0.0J": 544,
    "vWii-4.0.0U": 545,
    "vWii-4.0.0E": 546,
    "vWii-5.2.0J": 608,
    "vWii-5.2.0U": 609,
    "vWii-5.2.0E": 610,
}



# "title/__init__.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy

from .cert import *
from .content import *
from .crypto import *
from .iospatcher import *
from .nus import *
from .ticket import *
from .title import *
from .tmd import *
from .util import *
from .wad import *



# "nand/setting.py" from libWiiPy by NinjaCheetah & Contributors
# https://github.com/NinjaCheetah/libWiiPy
#
# See https://wiibrew.org/wiki//title/00000001/00000002/data/setting.txt for information about setting.txt.

import io
from ..shared import _pad_bytes


_key = 0x73B5DBFA

class SettingTxt:
    """
    A SettingTxt object that allows for decrypting and then parsing a setting.txt file from the Wii.

    Attributes
    ----------
    area : str
        The region of the System Menu this file matches with.
    model : str
        The model of the console, usually RVL-001 or RVL-101.
    dvd : int
        Unknown, might have to do with indicating support for scrapped DVD playback capabilities.
    mpch : str
        Unknown, generally accepted value is "0x7FFE".
    code : str
        Unknown code, may match with manufacturer code in serial number?
    serial_number : str
        Serial number of the console.
    video : str
        Video mode, either NTSC or PAL.
    game : str
        Another region code, possibly set by the hidden region select channel.
    """
    def __init__(self):
        self.area: str = ""
        self.model: str = ""
        self.dvd: int = 0
        self.mpch: str = ""  # What does this mean, Movie Player Channel? It's also a hex string, it seems.
        self.code: str = ""
        self.serial_number: str = ""
        self.video: str = ""
        self.game: str = ""

    def load(self, setting_txt: bytes) -> None:
        """
        Loads the raw data of an encrypted setting.txt file and decrypts it to parse its arguments

        Parameters
        ----------
        setting_txt : bytes
            The data of an encrypted setting.txt file.
        """
        with io.BytesIO(setting_txt) as setting_data:
            global _key  # I still don't actually know what *kind* of encryption this is.
            setting_txt_dec: [int] = []
            for i in range(0, 256):
                setting_txt_dec.append(int.from_bytes(setting_data.read(1)) ^ (_key & 0xff))
                _key = (_key << 1) | (_key >> 31)
        setting_txt_dec = bytes(setting_txt_dec)
        try:
            setting_str = setting_txt_dec.decode('utf-8')
        except UnicodeDecodeError:
            last_newline_pos = setting_txt_dec.rfind(b'\n')  # This makes sure we don't try to decode any garbage data.
            setting_str = setting_txt_dec[:last_newline_pos + 1].decode('utf-8')
        self.load_decrypted(setting_str)

    def load_decrypted(self, setting_txt: str) -> None:
        """
        Loads the raw data of a decrypted setting.txt file and parses its arguments

        Parameters
        ----------
        setting_txt : str
            The data of a decrypted setting.txt file.
        """
        setting_dict = {}
        # Iterate over every key in the file to create a dictionary.
        for line in setting_txt.splitlines():
            line = line.strip()
            if line is not None:
                key, value = line.split('=', 1)
                setting_dict[key.strip()] = value.strip()
        # Load the values from the dictionary into the object.
        self.area = setting_dict["AREA"]
        self.model = setting_dict["MODEL"]
        self.dvd = int(setting_dict["DVD"])
        self.mpch = setting_dict["MPCH"]
        self.code = setting_dict["CODE"]
        self.serial_number = setting_dict["SERNO"]
        self.video = setting_dict["VIDEO"]
        self.game = setting_dict["GAME"]

    def dump(self) -> bytes:
        """
        Dumps the SettingTxt object back into an encrypted bytes that the Wii can load.

        Returns
        -------
        bytes
            The setting.txt file as encrypted bytes.
        """
        setting_str = self.dump_decrypted()
        setting_txt_dec = setting_str.encode()
        global _key
        # This could probably be made more efficient somehow.
        setting_txt_enc: [int] = []
        with io.BytesIO(setting_txt_dec) as setting_data:
            for i in range(0, len(setting_txt_dec)):
                setting_txt_enc.append(int.from_bytes(setting_data.read(1)) ^ (_key & 0xff))
                _key = (_key << 1) | (_key >> 31)
        setting_txt_enc = _pad_bytes(bytes(setting_txt_enc), 256)
        return setting_txt_enc

    def dump_decrypted(self) -> str:
        """
        Dumps the SettingTxt object into a decrypted string.

        Returns
        -------
        str
            The setting.txt file as decrypted text.
        """
        # Write the keys back into a text file that can then be manually edited or re-encrypted.
        setting_txt = ""
        setting_txt += f"AREA={self.area}\r\n"
        setting_txt += f"MODEL={self.model}\r\n"
        setting_txt += f"DVD={self.dvd}\r\n"
        setting_txt += f"MPCH={self.mpch}\r\n"
        setting_txt += f"CODE={self.code}\r\n"
        setting_txt += f"SERNO={self.serial_number}\r\n"
        setting_txt += f"VIDEO={self.video}\r\n"
        setting_txt += f"GAME={self.game}\r\n"
        return setting_txt
